# pylint: disable=too-many-lines,no-member
import json
import re
from functools import partial
from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.lib.commands import cluster

from pcs_test.tier0.lib.commands.cluster.test_add_nodes import (
    QDEVICE_HOST,
    LocalConfig,
    corosync_conf_fixture,
    corosync_node_fixture,
    generate_nodes,
    node_fixture,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import patch_getaddrinfo

get_env_tools = partial(get_env_tools, local_extensions={"local": LocalConfig})


class GetTargets(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.existing_nodes_num = 3
        self.existing_nodes, self.new_nodes = generate_nodes(
            self.existing_nodes_num, 3
        )
        patch_getaddrinfo(self, self.new_nodes)
        self.existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        (
            self.config.local.set_expected_reports_list(self.expected_reports)
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(
                corosync_conf_fixture(self.existing_corosync_nodes)
            )
            .runner.cib.load()
        )

    def _add_nodes(self, skip_offline=False):
        force_flags = []
        if skip_offline:
            force_flags.append(reports.codes.SKIP_OFFLINE_NODES)
        cluster.add_nodes(
            self.env_assist.get_env(),
            [{"name": node, "addrs": [node]} for node in self.new_nodes],
            force_flags=force_flags,
        )

    def _add_nodes_with_lib_error(self, skip_offline=False):
        self.env_assist.assert_raise_library_error(
            lambda: self._add_nodes(skip_offline=skip_offline)
        )

    def test_some_existing_nodes_unknown(self):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes[1:] + self.new_nodes
            )
            .http.host.check_auth(node_labels=self.existing_nodes[1:])
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(self.new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self._add_nodes_with_lib_error()

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND,
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                    host_list=self.existing_nodes[:1],
                )
            ]
        )

    def test_some_existing_nodes_unknown_skipped(self):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes[1:] + self.new_nodes
            )
            .http.host.check_auth(node_labels=self.existing_nodes[1:])
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(self.new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
            .http.host.update_known_hosts(
                node_labels=self.new_nodes,
                to_add_hosts=self.existing_nodes[1:] + self.new_nodes,
            )
            .local.disable_sbd(self.new_nodes)
            .fs.isdir(settings.booth_config_dir, return_value=False)
            .local.no_file_sync()
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes
                    + [
                        node_fixture(node, i)
                        for i, node in enumerate(
                            self.new_nodes, self.existing_nodes_num + 1
                        )
                    ]
                ),
                self.existing_nodes[1:],
                self.new_nodes,
            )
        )

        self._add_nodes(skip_offline=True)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    reports.codes.HOST_NOT_FOUND,
                    host_list=self.existing_nodes[:1],
                )
            ]
        )

    def test_all_existing_nodes_unknown(self):
        (
            self.config.env.set_known_nodes(self.new_nodes)
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(self.new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self._add_nodes_with_lib_error()

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND,
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                    host_list=self.existing_nodes,
                ),
                fixture.error(reports.codes.NONE_HOST_FOUND),
            ]
        )

    def test_all_existing_nodes_unknown_skipped(self):
        (
            self.config.env.set_known_nodes(self.new_nodes)
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(self.new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self._add_nodes_with_lib_error(skip_offline=True)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    reports.codes.HOST_NOT_FOUND, host_list=self.existing_nodes
                ),
                fixture.error(reports.codes.NONE_HOST_FOUND),
            ]
        )

    def _assert_qnetd_unknown(self, skip_offline):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes + self.new_nodes
            )
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    self.existing_corosync_nodes, qdevice_net=True
                ),
                instead="corosync_conf.load_content",
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .local.get_host_info(self.new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self._add_nodes_with_lib_error(skip_offline=skip_offline)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=[QDEVICE_HOST]
                ),
            ]
        )

    def test_qnetd_unknown(self):
        self._assert_qnetd_unknown(False)

    def test_qnetd_unknown_skipped(self):
        self._assert_qnetd_unknown(True)

    def _assert_new_nodes_unknown(self, skip_offline):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes + self.new_nodes[1:]
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(self.new_nodes[1:])
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self._add_nodes_with_lib_error(skip_offline=skip_offline)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=self.new_nodes[:1]
                )
            ]
        )

    def test_new_nodes_unknown(self):
        self._assert_new_nodes_unknown(False)

    def test_new_nodes_unknown_skipped(self):
        self._assert_new_nodes_unknown(True)


class NoneNamesMissing(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        existing_nodes_num = 3
        self.existing_nodes, self.new_nodes = generate_nodes(
            existing_nodes_num, 3
        )
        patch_getaddrinfo(self, self.new_nodes)
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.corosync_conf = corosync_conf_fixture(existing_corosync_nodes)
        self.existing_nodes_with_name = self.existing_nodes

    def _add_nodes_with_lib_error(self, corosync_conf):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes + self.new_nodes
            )
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(corosync_conf)
            .runner.cib.load()
        )
        if self.existing_nodes_with_name:
            self.config.http.host.check_auth(
                node_labels=self.existing_nodes_with_name
            )

        self.config.services.is_installed("sbd", return_value=False)
        self.config.local.get_host_info(self.new_nodes)
        self.config.local.pcsd_ssl_cert_sync_disabled()

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": node, "addrs": [node]} for node in self.new_nodes],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_some_node_names_missing(self):
        corosync_conf = re.sub(r"\s+name: node1\n", "\n", self.corosync_conf)
        # make sure the name was removed
        self.assertNotEqual(corosync_conf, self.corosync_conf)
        self.existing_nodes_with_name = self.existing_nodes[1:]
        self._add_nodes_with_lib_error(corosync_conf)

    def test_all_node_names_missing(self):
        corosync_conf = re.sub(r"\s+name: .*\n", "\n", self.corosync_conf)
        # make sure the names were removed
        self.assertNotEqual(corosync_conf, self.corosync_conf)
        self.existing_nodes_with_name = []
        self._add_nodes_with_lib_error(corosync_conf)


class Inputs(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_conflict_existing_nodes(self):
        existing_nodes = ["node1", "node2", "node3"]
        new_nodes = ["new1", "remote-name", "node3", "guest-name"]
        node1_addrs = ["new-addr1", "addr1-2", "guest-host", "remote-host"]
        patch_getaddrinfo(self, new_nodes + node1_addrs)

        self.config.env.set_known_nodes(existing_nodes + new_nodes)
        self.config.services.is_enabled("sbd", return_value=False)
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(
                [
                    corosync_node_fixture(1, "node1", ["addr1-1", "addr1-2"]),
                    corosync_node_fixture(2, "node2", ["addr2-1", "addr2-2"]),
                    corosync_node_fixture(3, "node3", ["addr3-1", "addr3-2"]),
                ]
            )
        )
        self.config.runner.cib.load(
            resources="""
                    <resources>
                        <primitive id="guest" class="ocf" provider="heartbeat"
                            type="VirtualDomain"
                        >
                            <meta_attributes id="guest-meta_attributes">
                                <nvpair id="guest-remote-addr"
                                    name="remote-addr" value="guest-host"
                                />
                                <nvpair id="guest-remote-node"
                                    name="remote-node" value="guest-name"
                                />
                             </meta_attributes>
                        </primitive>
                        <primitive id="remote-name" class="ocf"
                            provider="pacemaker" type="remote"
                        >
                            <instance_attributes id="remote-instance_attrs">
                                <nvpair id="remote-server"
                                    name="server" value="remote-host"
                                />
                            </instance_attributes>
                        </primitive>
                    </resources>
                """
        )
        self.config.http.host.check_auth(node_labels=existing_nodes)
        self.config.local.get_host_info(new_nodes)
        self.config.local.pcsd_ssl_cert_sync_disabled()

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    # no change, addrs defined
                    {"name": new_nodes[0], "addrs": node1_addrs},
                    # no change, addrs defined even though empty
                    {"name": new_nodes[1], "addrs": []},
                    # use a default address
                    {"name": new_nodes[2], "addrs": None},
                    # use a default address
                    {"name": new_nodes[3]},
                ],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name="node3",
                    address="node3",
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                ),
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name="guest-name",
                    address="guest-name",
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                ),
                fixture.error(
                    reports.codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=4,
                    min_count=2,
                    max_count=2,
                    node_name="new1",
                    node_index=1,
                ),
                fixture.error(
                    reports.codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=0,
                    min_count=2,
                    max_count=2,
                    node_name="remote-name",
                    node_index=2,
                ),
                fixture.error(
                    reports.codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=1,
                    min_count=2,
                    max_count=2,
                    node_name="node3",
                    node_index=3,
                ),
                fixture.error(
                    reports.codes.COROSYNC_BAD_NODE_ADDRESSES_COUNT,
                    actual_count=1,
                    min_count=2,
                    max_count=2,
                    node_name="guest-name",
                    node_index=4,
                ),
                fixture.error(
                    reports.codes.NODE_NAMES_ALREADY_EXIST,
                    name_list=["guest-name", "node3", "remote-name"],
                ),
                fixture.error(
                    reports.codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["addr1-2", "guest-host", "remote-host"],
                ),
            ]
        )

    def conflict_existing_nodes_cib_load_error(self):
        existing_nodes = ["node1", "node2", "node3", "node4"]
        new_nodes = ["new1"]

        self.config.env.set_known_nodes(existing_nodes + new_nodes)
        self.config.services.is_enabled("sbd", return_value=False)
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(
                [
                    corosync_node_fixture(1, "node1", ["addr1-1"]),
                    corosync_node_fixture(2, "node2", ["addr2-1"]),
                    corosync_node_fixture(3, "node3", ["addr3-1"]),
                    corosync_node_fixture(4, "node4", ["addr4-1"]),
                ]
            )
        )
        self.config.runner.cib.load(returncode=1, stderr="an error")
        self.config.http.host.check_auth(node_labels=existing_nodes)
        self.config.local.get_host_info(new_nodes)
        self.config.local.pcsd_ssl_cert_sync_disabled()

    def test_conflict_existing_nodes_cib_load_error(self):
        node_name = "new1"
        patch_getaddrinfo(self, [node_name])
        self.conflict_existing_nodes_cib_load_error()
        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": node_name}],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node_name,
                    address=node_name,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                ),
                fixture.error(
                    reports.codes.CIB_LOAD_ERROR_GET_NODES_FOR_VALIDATION,
                    force_code=reports.codes.FORCE,
                ),
            ]
        )

    def test_conflict_existing_nodes_cib_load_error_forced(self):
        node_addrs = ["addr1-1"]
        patch_getaddrinfo(self, node_addrs)
        self.conflict_existing_nodes_cib_load_error()
        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1", "addrs": node_addrs}],
                force_flags=[reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.CIB_LOAD_ERROR_GET_NODES_FOR_VALIDATION
                ),
                fixture.error(
                    reports.codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=node_addrs,
                ),
            ]
        )

    def test_force_unresolvable(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        patch_getaddrinfo(self, [])
        self.config.env.set_known_nodes(existing_nodes + new_nodes)
        self.config.services.is_enabled("sbd", return_value=False)
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(
                [
                    node_fixture(node, i)
                    for i, node in enumerate(existing_nodes, 1)
                ]
            )
        )
        self.config.runner.cib.load()
        self.config.http.host.check_auth(node_labels=existing_nodes)
        self.config.local.get_host_info(new_nodes)
        self.config.local.pcsd_ssl_cert_sync_disabled()

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    # Use an existing address so the command stops after the
                    # validation and the test does not have to cover the whole
                    # node add process.
                    {"name": "new1", "addrs": ["node1"]},
                ],
                force_flags=[reports.codes.FORCE],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=["node1"],
                ),
                fixture.error(
                    reports.codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["node1"],
                ),
            ]
        )

    def test_sbd_disabled(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1", "new2", "new3", "new4"]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .http.host.check_auth(node_labels=existing_nodes)
            .services.is_installed("sbd", return_value=False)
            .local.get_host_info(new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    {
                        "name": "new1",
                        "watchdog": "/dev/wd",
                        "devices": ["/dev/sxa", "/dev/sxb"],
                    },
                    {"name": "new2", "devices": ["/dev/sxc"]},
                    {"name": "new3", "watchdog": "/dev/wd"},
                    {"name": "new4", "nonsense": "option"},
                ],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["nonsense"],
                    option_type="node",
                    allowed=["addrs", "devices", "name", "watchdog"],
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.SBD_NOT_USED_CANNOT_SET_SBD_OPTIONS,
                    node="new1",
                    options=["devices", "watchdog"],
                ),
                fixture.error(
                    reports.codes.SBD_NOT_USED_CANNOT_SET_SBD_OPTIONS,
                    node="new2",
                    options=["devices"],
                ),
                fixture.error(
                    reports.codes.SBD_NOT_USED_CANNOT_SET_SBD_OPTIONS,
                    node="new3",
                    options=["watchdog"],
                ),
            ]
        )

    def test_sbd_enabled_without_device(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1", "new2", "new3"]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=True)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .local.read_sbd_config()
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .http.sbd.check_sbd(
                # This is where the values validation happens. This test only
                # deals with input params validation happening on the local
                # node, so we pretend here the remote nodes say everything is
                # valid
                communication_list=[
                    fixture.check_sbd_comm_success_fixture(
                        "new1", "/dev/wd", ["/dev/sxa", "/dev/sxb"]
                    ),
                    fixture.check_sbd_comm_success_fixture(
                        "new2", "/dev/watchdog", ["/dev/sxc"]
                    ),
                    fixture.check_sbd_comm_success_fixture(
                        "new3", "/dev/wdog", []
                    ),
                ]
            )
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    {
                        "name": "new1",
                        "watchdog": "/dev/wd",
                        "devices": ["/dev/sxa", "/dev/sxb"],
                        "nonsense": "option",
                    },
                    {"name": "new2", "devices": ["/dev/sxc"]},
                    {"name": "new3", "watchdog": "/dev/wdog"},
                ],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.info(
                    reports.codes.USING_DEFAULT_WATCHDOG,
                    node="new2",
                    watchdog="/dev/watchdog",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["nonsense"],
                    option_type="node",
                    allowed=["addrs", "devices", "name", "watchdog"],
                    allowed_patterns=[],
                ),
            ]
            + [
                fixture.error(
                    reports.codes.SBD_WITH_DEVICES_NOT_USED_CANNOT_SET_DEVICE,
                    node=node,
                )
                for node in ["new1", "new2"]
            ]
            + [fixture.info(reports.codes.SBD_CHECK_STARTED)]
            + [
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node=node)
                for node in new_nodes
            ]
        )

    def test_sbd_enabled_with_device(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1", "new2", "new3"]
        patch_getaddrinfo(self, new_nodes)
        devices1 = ["/dev/sxa", "/dev/sxb", "/dev/sxc", "/dev/sxd"]
        devices2 = ["/dev/sxe", "dev/sxf"]
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=True)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .local.read_sbd_config("SBD_DEVICE=/device\n")
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .http.sbd.check_sbd(
                # This is where the values validation happens. This test only
                # deals with input params validation happening on the local
                # node, so we pretend here the remote nodes say everything is
                # valid
                communication_list=[
                    fixture.check_sbd_comm_success_fixture(
                        "new1", "/dev/wd", devices1
                    ),
                    fixture.check_sbd_comm_success_fixture(
                        "new2", "/dev/watchdog", devices2
                    ),
                    fixture.check_sbd_comm_success_fixture(
                        "new3", "/dev/wdog", []
                    ),
                ]
            )
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    {
                        "name": "new1",
                        "watchdog": "/dev/wd",
                        "devices": devices1,
                        "nonsense": "option",
                    },
                    {"name": "new2", "devices": devices2},
                    {"name": "new3", "watchdog": "/dev/wdog"},
                ],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.info(
                    reports.codes.USING_DEFAULT_WATCHDOG,
                    node="new2",
                    watchdog="/dev/watchdog",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["nonsense"],
                    option_type="node",
                    allowed=["addrs", "devices", "name", "watchdog"],
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.SBD_TOO_MANY_DEVICES_FOR_NODE,
                    node="new1",
                    device_list=devices1,
                    max_devices=3,
                ),
                fixture.error(
                    reports.codes.SBD_DEVICE_PATH_NOT_ABSOLUTE,
                    node="new2",
                    device="dev/sxf",
                ),
                fixture.error(
                    reports.codes.SBD_NO_DEVICE_FOR_NODE,
                    node="new3",
                    sbd_enabled_in_cluster=True,
                ),
                fixture.info(reports.codes.SBD_CHECK_STARTED),
            ]
            + [
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node=node)
                for node in new_nodes
            ]
        )

    def test_wait_without_start(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(), [{"name": "new1"}], wait=10
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [fixture.error(reports.codes.WAIT_FOR_NODE_STARTUP_WITHOUT_START)]
        )

    def test_wait(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1"}],
                start=True,
                wait="nonsense",
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.INVALID_TIMEOUT_VALUE, timeout="nonsense"
                )
            ]
        )


class ClusterStatus(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def setup_config(
        self,
        existing_nodes,
        new_nodes,
        check_auth_communication_list=None,
        with_get_host_info=True,
    ):
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=False)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
        )
        if check_auth_communication_list:
            self.config.http.host.check_auth(
                communication_list=check_auth_communication_list
            )
        else:
            self.config.http.host.check_auth(node_labels=existing_nodes)
        if with_get_host_info:
            self.config.local.get_host_info(new_nodes)
            self.config.local.pcsd_ssl_cert_sync_disabled()

    def test_all_nodes_offline_skipped(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        self.setup_config(
            existing_nodes,
            new_nodes,
            [
                {
                    "label": "node1",
                    "was_connected": False,
                    "errno": 7,
                    "error_msg": "an error",
                },
                {
                    "label": "node2",
                    "response_code": 400,
                    "output": "not authorized",
                },
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1"}],
                force_flags=[reports.codes.SKIP_OFFLINE_NODES],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.warn(
                    reports.codes.OMITTING_NODE,
                    node="node1",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="node2",
                    command="remote/check_auth",
                    reason="not authorized",
                ),
                fixture.error(
                    reports.codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE
                ),
            ]
        )

    def test_some_nodes_offline(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        self.setup_config(
            existing_nodes,
            new_nodes,
            [
                {
                    "label": "node1",
                    "was_connected": False,
                    "errno": 7,
                    "error_msg": "an error",
                },
                {
                    "label": "node2",
                    "response_code": 200,
                    "output": '{"success":true}',
                },
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1"}],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                    node="node1",
                    command="remote/check_auth",
                    reason="an error",
                ),
            ]
        )

    def test_some_nodes_offline_skipped(self):
        existing_nodes = ["node1", "node2"]
        new_nodes = ["new1"]
        self.setup_config(
            existing_nodes,
            new_nodes,
            [
                {
                    "label": "node1",
                    "was_connected": False,
                    "errno": 7,
                    "error_msg": "an error",
                },
                {
                    "label": "node2",
                    "response_code": 200,
                    "output": '{"success":true}',
                },
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1"}],
                # Use 'wait' without 'start' so the command stops after the
                # validation and the test does not have to cover the whole
                # node add process.
                wait=10,
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.WAIT_FOR_NODE_STARTUP_WITHOUT_START
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                    node="node1",
                    command="remote/check_auth",
                    reason="an error",
                ),
            ]
        )

    def test_atb_will_be_enable_cluster_not_offline(self):
        existing_nodes = ["node1", "node2", "node3", "node4", "node5"]
        new_nodes = ["new1"]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled(
                "sbd", return_value=True, name="is_enabled_sbd_1"
            )
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .local.read_sbd_config(name_sufix="_1")
            .http.host.check_auth(node_labels=existing_nodes)
            .services.is_installed("sbd", return_value=True)
            .services.is_enabled(
                "sbd", return_value=True, name="is_enabled_sbd_2"
            )
            .local.read_sbd_config(name_sufix="_2")
            .http.corosync.check_corosync_offline(
                communication_list=[
                    {"label": "node1", "output": '{"corosync":true}'},
                    {"label": "node2", "output": "an error"},
                    {
                        "label": "node3",
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {"label": "node4", "output": '{"corosync":true}'},
                    {"label": "node5", "output": '{"corosync":false}'},
                ]
            )
            .local.get_host_info(new_nodes)
            .http.sbd.check_sbd(
                # This is where the SBD values validation happens. This test
                # only deals with enabling ATB check, so we pretend here the
                # remote nodes say everything is valid
                communication_list=[
                    fixture.check_sbd_comm_success_fixture(
                        "new1", "/dev/watchdog", []
                    ),
                ]
            )
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": "new1"}],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.info(
                    reports.codes.USING_DEFAULT_WATCHDOG,
                    node="new1",
                    watchdog="/dev/watchdog",
                ),
                fixture.warn(
                    reports.codes.COROSYNC_QUORUM_ATB_WILL_BE_ENABLED_DUE_TO_SBD
                ),
                fixture.info(reports.codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
                fixture.error(
                    reports.codes.COROSYNC_RUNNING_ON_NODE, node="node1"
                ),
                fixture.error(
                    reports.codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node="node2",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="node3",
                    command="remote/status",
                    reason="an error",
                ),
                fixture.error(
                    reports.codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node="node3",
                ),
                fixture.error(
                    reports.codes.COROSYNC_RUNNING_ON_NODE, node="node4"
                ),
                fixture.info(
                    reports.codes.COROSYNC_NOT_RUNNING_ON_NODE, node="node5"
                ),
                fixture.info(reports.codes.SBD_CHECK_STARTED),
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node="new1"),
            ]
        )

    @staticmethod
    def fixture_get_host_info_communication():
        return [
            dict(
                label="new1",
                output=json.dumps(
                    dict(
                        services=dict(
                            corosync=dict(
                                installed=True, enabled=True, running=True
                            ),
                            pacemaker=dict(
                                installed=True, enabled=True, running=True
                            ),
                            pcsd=dict(
                                installed=True, enabled=True, running=True
                            ),
                        ),
                        cluster_configuration_exists=False,
                    )
                ),
            ),
            dict(
                label="new2",
                output=json.dumps(
                    dict(
                        services=dict(
                            corosync=dict(
                                installed=False, enabled=False, running=False
                            ),
                            pacemaker=dict(
                                installed=False, enabled=False, running=False
                            ),
                            pcsd=dict(
                                installed=True, enabled=True, running=True
                            ),
                        ),
                        cluster_configuration_exists=False,
                    )
                ),
            ),
            dict(
                label="new3",
                output=json.dumps(
                    dict(
                        services=dict(
                            corosync=dict(
                                installed=True, enabled=True, running=False
                            ),
                            pacemaker=dict(
                                installed=True, enabled=True, running=False
                            ),
                            pcsd=dict(
                                installed=True, enabled=True, running=True
                            ),
                        ),
                        cluster_configuration_exists=True,
                    )
                ),
            ),
            dict(label="new4", output=json.dumps({})),
            dict(
                label="new5", was_connected=False, errno=7, error_msg="an error"
            ),
            dict(label="new6", response_code=400, output="an error"),
        ]

    def test_new_nodes_not_ready(self):
        existing_nodes = ["node1", "node2", "node3"]
        new_nodes = ["new1", "new2", "new3", "new4", "new5", "new6"]
        self.setup_config(existing_nodes, new_nodes, with_get_host_info=False)
        self.config.http.host.get_host_info(
            communication_list=self.fixture_get_host_info_communication()
        )
        self.config.local.pcsd_ssl_cert_sync_disabled()

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": name} for name in new_nodes],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="new5",
                    command="remote/check_host",
                    reason="an error",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="new6",
                    command="remote/check_host",
                    reason="an error",
                ),
                fixture.error(
                    reports.codes.HOST_ALREADY_IN_CLUSTER_SERVICES,
                    host_name="new1",
                    service_list=["corosync", "pacemaker"],
                    force_code=reports.codes.FORCE,
                ),
                fixture.error(
                    reports.codes.SERVICE_NOT_INSTALLED,
                    node="new2",
                    service_list=["corosync", "pacemaker"],
                ),
                fixture.error(
                    reports.codes.HOST_ALREADY_IN_CLUSTER_CONFIG,
                    host_name="new3",
                    force_code=reports.codes.FORCE,
                ),
                fixture.error(
                    reports.codes.INVALID_RESPONSE_FORMAT,
                    node="new4",
                ),
                fixture.error(
                    reports.codes.CLUSTER_WILL_BE_DESTROYED,
                    force_code=reports.codes.FORCE,
                ),
            ]
        )

    def test_new_nodes_not_ready_forced(self):
        existing_nodes = ["node1", "node2", "node3"]
        new_nodes = ["new1", "new2", "new3", "new4", "new5", "new6"]
        self.setup_config(existing_nodes, new_nodes, with_get_host_info=False)
        self.config.http.host.get_host_info(
            communication_list=self.fixture_get_host_info_communication()
        )
        self.config.local.pcsd_ssl_cert_sync_disabled()

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [{"name": name} for name in new_nodes],
                force_flags=[reports.codes.FORCE],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="new5",
                    command="remote/check_host",
                    reason="an error",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="new6",
                    command="remote/check_host",
                    reason="an error",
                ),
                fixture.warn(
                    reports.codes.HOST_ALREADY_IN_CLUSTER_SERVICES,
                    host_name="new1",
                    service_list=["corosync", "pacemaker"],
                ),
                fixture.error(
                    reports.codes.SERVICE_NOT_INSTALLED,
                    node="new2",
                    service_list=["corosync", "pacemaker"],
                ),
                fixture.warn(
                    reports.codes.HOST_ALREADY_IN_CLUSTER_CONFIG,
                    host_name="new3",
                ),
                fixture.error(
                    reports.codes.INVALID_RESPONSE_FORMAT,
                    node="new4",
                ),
            ]
        )

    @staticmethod
    def fixture_sbd_check_input(suffix, has_wd=True):
        return [
            ("watchdog", f"/dev/watchdog{suffix}" if has_wd else ""),
            ("device_list", json.dumps([f"/dev/sda{suffix}"])),
        ]

    @staticmethod
    def fixture_sbd_check_output(
        suffix,
        sbd_installed=True,
        wd_exists=True,
        wd_is_supported=True,
        device_exists=True,
        device_block=True,
        has_wd=True,
    ):
        result = {
            "sbd": {
                "installed": sbd_installed,
            },
            "watchdog": {
                "exist": wd_exists,
                "path": f"/dev/watchdog{suffix}",
                "is_supported": wd_is_supported,
            },
            "device_list": [
                {
                    "path": f"/dev/sda{suffix}",
                    "exist": device_exists,
                    "block_device": device_block,
                },
            ],
        }
        if has_wd:
            result["watchdog"] = {
                "exist": wd_exists,
                "path": f"/dev/watchdog{suffix}",
                "is_supported": wd_is_supported,
            }
        return json.dumps(result)

    def test_sbd_check(self):
        existing_nodes = ["node1", "node2", "node3", "node4"]
        new_nodes = [f"new{i}" for i in range(1, 10)]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=True)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .local.read_sbd_config("SBD_DEVICE=/device\n")
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .http.sbd.check_sbd(
                communication_list=[
                    {
                        "label": "new1",
                        "output": self.fixture_sbd_check_output(
                            1, sbd_installed=False
                        ),
                        "param_list": self.fixture_sbd_check_input(1),
                    },
                    {
                        "label": "new2",
                        "output": self.fixture_sbd_check_output(
                            2,
                            wd_exists=False,
                            wd_is_supported=False,
                        ),
                        "param_list": self.fixture_sbd_check_input(2),
                    },
                    {
                        "label": "new3",
                        "output": self.fixture_sbd_check_output(
                            3, device_exists=False, device_block=False
                        ),
                        "param_list": self.fixture_sbd_check_input(3),
                    },
                    {
                        "label": "new4",
                        "output": self.fixture_sbd_check_output(
                            4, device_block=False
                        ),
                        "param_list": self.fixture_sbd_check_input(4),
                    },
                    {
                        "label": "new5",
                        "output": "bad json",
                        "param_list": self.fixture_sbd_check_input(5),
                    },
                    {
                        "label": "new6",
                        "response_code": 400,
                        "output": "an error",
                        "param_list": self.fixture_sbd_check_input(6),
                    },
                    {
                        "label": "new7",
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                        "param_list": self.fixture_sbd_check_input(7),
                    },
                    {
                        "label": "new8",
                        "output": self.fixture_sbd_check_output(8),
                        "param_list": self.fixture_sbd_check_input(8),
                    },
                    {
                        "label": "new9",
                        "output": self.fixture_sbd_check_output(
                            9,
                            wd_is_supported=False,
                        ),
                        "param_list": self.fixture_sbd_check_input(9),
                    },
                ]
            )
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    {
                        "name": f"new{i}",
                        "watchdog": f"/dev/watchdog{i}",
                        "devices": [f"/dev/sda{i}"],
                    }
                    for i in range(1, 10)
                ],
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.info(reports.codes.SBD_CHECK_STARTED),
                fixture.error(reports.codes.SBD_NOT_INSTALLED, node="new1"),
                fixture.error(
                    reports.codes.WATCHDOG_NOT_FOUND,
                    node="new2",
                    watchdog="/dev/watchdog2",
                ),
                fixture.error(
                    reports.codes.SBD_DEVICE_DOES_NOT_EXIST,
                    node="new3",
                    device="/dev/sda3",
                ),
                fixture.error(
                    reports.codes.SBD_DEVICE_IS_NOT_BLOCK_DEVICE,
                    node="new4",
                    device="/dev/sda4",
                ),
                fixture.error(
                    reports.codes.INVALID_RESPONSE_FORMAT,
                    node="new5",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="new6",
                    command="remote/check_sbd",
                    reason="an error",
                ),
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node="new7",
                    command="remote/check_sbd",
                    reason="an error",
                ),
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node="new8"),
                fixture.error(
                    reports.codes.SBD_WATCHDOG_NOT_SUPPORTED,
                    node="new9",
                    watchdog="/dev/watchdog9",
                ),
            ]
        )

    def test_sbd_check_forced(self):
        existing_nodes = ["node1", "node2", "node3", "node4"]
        new_nodes = [f"new{i}" for i in range(1, 4)]
        patch_getaddrinfo(self, new_nodes)
        (
            self.config.env.set_known_nodes(existing_nodes + new_nodes)
            .services.is_enabled("sbd", return_value=True)
            .corosync_conf.load_content(
                corosync_conf_fixture(
                    [
                        node_fixture(node, i)
                        for i, node in enumerate(existing_nodes, 1)
                    ]
                )
            )
            .runner.cib.load()
            .local.read_sbd_config("SBD_DEVICE=/device\n")
            .http.host.check_auth(node_labels=existing_nodes)
            .local.get_host_info(new_nodes)
            .http.sbd.check_sbd(
                communication_list=[
                    {
                        "label": "new1",
                        "output": self.fixture_sbd_check_output(
                            1,
                            sbd_installed=False,
                            has_wd=False,
                        ),
                        "param_list": self.fixture_sbd_check_input(
                            1,
                            has_wd=False,
                        ),
                    },
                    {
                        "label": "new2",
                        "output": self.fixture_sbd_check_output(
                            2,
                            has_wd=False,
                        ),
                        "param_list": self.fixture_sbd_check_input(
                            2,
                            has_wd=False,
                        ),
                    },
                    {
                        "label": "new3",
                        "output": self.fixture_sbd_check_output(
                            3,
                            has_wd=False,
                        ),
                        "param_list": self.fixture_sbd_check_input(
                            3,
                            has_wd=False,
                        ),
                    },
                ]
            )
            .local.pcsd_ssl_cert_sync_disabled()
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.add_nodes(
                self.env_assist.get_env(),
                [
                    {
                        "name": f"new{i}",
                        "watchdog": f"/dev/watchdog{i}",
                        "devices": [f"/dev/sda{i}"],
                    }
                    for i in range(1, 4)
                ],
                no_watchdog_validation=True,
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.USING_DEFAULT_ADDRESS_FOR_HOST,
                    host_name=node,
                    address=node,
                    address_source=(
                        reports.const.DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS
                    ),
                )
                for node in new_nodes
            ]
            + [
                fixture.info(reports.codes.SBD_CHECK_STARTED),
                fixture.error(reports.codes.SBD_NOT_INSTALLED, node="new1"),
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node="new2"),
                fixture.info(reports.codes.SBD_CHECK_SUCCESS, node="new3"),
                fixture.warn(reports.codes.SBD_WATCHDOG_VALIDATION_INACTIVE),
            ]
        )



from pcs_test.tools.command_env.config_runner_booth import BoothShortcuts
from pcs_test.tools.command_env.config_runner_cib import CibShortcuts
from pcs_test.tools.command_env.config_runner_corosync import CorosyncShortcuts
from pcs_test.tools.command_env.config_runner_pcmk import PcmkShortcuts
from pcs_test.tools.command_env.config_runner_sbd import SbdShortcuts
from pcs_test.tools.command_env.config_runner_scsi import ScsiShortcuts
from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class RunnerConfig:
    def __init__(self, call_collection, wrap_helper):
        self.__calls = call_collection

        self.booth = wrap_helper(BoothShortcuts(self.__calls))
        self.cib = wrap_helper(CibShortcuts(self.__calls))
        self.corosync = wrap_helper(CorosyncShortcuts(self.__calls))
        self.pcmk = wrap_helper(PcmkShortcuts(self.__calls))
        self.sbd = wrap_helper(SbdShortcuts(self.__calls))
        self.scsi = wrap_helper(ScsiShortcuts(self.__calls))

    def place(
        self,
        command,
        name="",
        stdout="",
        stderr="",
        returncode=0,
        check_stdin=None,
        before=None,
        instead=None,
        env=None,
    ):
        # pylint: disable=too-many-arguments
        """
        Place new call to a config.

        string command -- cmdline call (e.g. "crm_mon --one-shot --as-xml")
        string name -- name of the call; it is possible to get it by the method
            "get"
        string stdout -- stdout of the call
        string stderr -- stderr of the call
        int returncode -- returncode of the call
        callable check_stdin -- callable that can check if stdin is as expected
        string before -- name of another call to insert this call before it
        string instead -- name of another call to replace it by this call
        dict env -- CommandRunner environment variables
        """
        call = RunnerCall(
            command, stdout, stderr, returncode, check_stdin, env=env
        )
        self.__calls.place(name, call, before, instead)
        return self



import sys
import xml.dom.minidom
from collections import defaultdict
from enum import Enum
from os.path import isfile
from xml.dom.minidom import parseString

import pcs.cli.constraint_colocation.command as colocation_command
import pcs.cli.constraint_order.command as order_command
from pcs import rule as rule_utils
from pcs import (
    settings,
    usage,
    utils,
)
from pcs.cli.common import parse_args
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint_ticket import command as ticket_command
from pcs.cli.reports import process_library_reports
from pcs.cli.reports.output import warn
from pcs.common import (
    const,
    pacemaker,
    reports,
)
from pcs.common.reports import ReportItem
from pcs.common.reports.constraints import colocation as colocation_format
from pcs.common.reports.constraints import order as order_format
from pcs.common.str_tools import format_list
from pcs.lib.cib.constraint import resource_set
from pcs.lib.cib.constraint.order import ATTRIB as order_attrib
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pacemaker.values import (
    SCORE_INFINITY,
    sanitize_id,
)

# pylint: disable=too-many-branches, too-many-statements
# pylint: disable=invalid-name, too-many-nested-blocks
# pylint: disable=too-many-locals, too-many-lines

OPTIONS_ACTION = resource_set.ATTRIB["action"]

DEFAULT_ACTION = "start"
DEFAULT_ROLE = const.PCMK_ROLE_STARTED

OPTIONS_SYMMETRICAL = order_attrib["symmetrical"]
OPTIONS_KIND = order_attrib["kind"]

LOCATION_NODE_VALIDATION_SKIP_MSG = (
    "Validation for node existence in the cluster will be skipped"
)
CRM_RULE_MISSING_MSG = (
    "Warning: crm_rule is not available, therefore expired constraints may be "
    "shown. Consider upgrading pacemaker.\n"
)

RESOURCE_TYPE_RESOURCE = "resource"
RESOURCE_TYPE_REGEXP = "regexp"

RULE_IN_EFFECT = "in effect"
RULE_EXPIRED = "expired"
RULE_NOT_IN_EFFECT = "not yet in effect"
RULE_UNKNOWN_STATUS = "unknown status"


class CrmRuleReturnCode(Enum):
    IN_EFFECT = 0
    EXPIRED = 110
    TO_BE_IN_EFFECT = 111


def constraint_location_cmd(lib, argv, modifiers):
    if not argv:
        sub_cmd = "config"
    else:
        sub_cmd = argv.pop(0)

    try:
        if sub_cmd == "add":
            location_add(lib, argv, modifiers)
        elif sub_cmd in ["remove", "delete"]:
            location_remove(lib, argv, modifiers)
        elif sub_cmd == "show":
            location_show(lib, argv, modifiers)
        elif sub_cmd == "config":
            location_config_cmd(lib, argv, modifiers)
        elif len(argv) >= 2:
            if argv[0] == "rule":
                location_rule(lib, [sub_cmd] + argv, modifiers)
            else:
                location_prefer(lib, [sub_cmd] + argv, modifiers)
        else:
            raise CmdLineInputError()
    except CmdLineInputError as e:
        utils.exit_on_cmdline_input_error(
            e, "constraint", ["location", sub_cmd]
        )


def constraint_order_cmd(lib, argv, modifiers):
    if not argv:
        sub_cmd = "config"
    else:
        sub_cmd = argv.pop(0)

    try:
        if sub_cmd == "set":
            order_command.create_with_set(lib, argv, modifiers)
        elif sub_cmd in ["remove", "delete"]:
            order_rm(lib, argv, modifiers)
        elif sub_cmd == "show":
            order_command.show(lib, argv, modifiers)
        elif sub_cmd == "config":
            order_command.config_cmd(lib, argv, modifiers)
        else:
            order_start(lib, [sub_cmd] + argv, modifiers)
    except CmdLineInputError as e:
        utils.exit_on_cmdline_input_error(e, "constraint", ["order", sub_cmd])


def constraint_show(lib, argv, modifiers):
    warn(
        "This command is deprecated and will be removed. "
        "Please use 'pcs constraint config' instead.",
        stderr=True,
    )
    return constraint_config_cmd(lib, argv, modifiers)


def constraint_config_cmd(lib, argv, modifiers):
    """
    Options:
      * --all - print expired constraints
      * -f - CIB file
      * --full
    """
    location_config_cmd(lib, argv, modifiers)
    order_command.config_cmd(lib, argv, modifiers.get_subset("--full", "-f"))
    colocation_command.config_cmd(
        lib, argv, modifiers.get_subset("--full", "-f")
    )
    ticket_command.config_cmd(lib, argv, modifiers.get_subset("--full", "-f"))


def colocation_rm(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    elementFound = False
    if len(argv) < 2:
        raise CmdLineInputError()

    (dom, constraintsElement) = getCurrentConstraints()

    resource1 = argv[0]
    resource2 = argv[1]

    for co_loc in constraintsElement.getElementsByTagName("rsc_colocation")[:]:
        if (
            co_loc.getAttribute("rsc") == resource1
            and co_loc.getAttribute("with-rsc") == resource2
        ):
            constraintsElement.removeChild(co_loc)
            elementFound = True
        if (
            co_loc.getAttribute("rsc") == resource2
            and co_loc.getAttribute("with-rsc") == resource1
        ):
            constraintsElement.removeChild(co_loc)
            elementFound = True

    if elementFound:
        utils.replace_cib_configuration(dom)
    else:
        print("No matching resources found in ordering list")


def _validate_constraint_resource(cib_dom, resource_id):
    (
        resource_valid,
        resource_error,
        dummy_correct_id,
    ) = utils.validate_constraint_resource(cib_dom, resource_id)
    if not resource_valid:
        utils.err(resource_error)


def _validate_resources_not_in_same_group(cib_dom, resource1, resource2):
    if not utils.validate_resources_not_in_same_group(
        cib_dom, resource1, resource2
    ):
        utils.err(
            "Cannot create an order constraint for resources in the same group"
        )


# Syntax: colocation add [role] <src> with [role] <tgt> [score] [options]
# possible commands:
#        <src> with        <tgt> [score] [options]
#        <src> with <role> <tgt> [score] [options]
# <role> <src> with        <tgt> [score] [options]
# <role> <src> with <role> <tgt> [score] [options]
def colocation_add(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --force - allow constraint on any resource, allow duplicate constraints
    """

    def _parse_score_options(argv):
        # When passed an array of arguments if the first argument doesn't have
        # an '=' then it's the score, otherwise they're all arguments. Return a
        # tuple with the score and array of name,value pairs
        """
        Commandline options: no options
        """
        if not argv:
            return SCORE_INFINITY, []
        score = SCORE_INFINITY if "=" in argv[0] else argv.pop(0)
        # create a list of 2-tuples (name, value)
        arg_array = [
            parse_args.split_option(arg, allow_empty_value=False)
            for arg in argv
        ]
        return score, arg_array

    del lib
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    role1 = ""
    role2 = ""

    cib_dom = utils.get_cib_dom()
    new_roles_supported = utils.isCibVersionSatisfied(
        cib_dom, const.PCMK_NEW_ROLES_CIB_VERSION
    )

    def _validate_and_prepare_role(role):
        role_cleaned = role.lower().capitalize()
        if role_cleaned not in const.PCMK_ROLES:
            utils.err(
                "invalid role value '{0}', allowed values are: {1}".format(
                    role, format_list(const.PCMK_ROLES)
                )
            )
        return pacemaker.role.get_value_for_cib(
            role_cleaned, new_roles_supported
        )

    if argv[2] == "with":
        role1 = _validate_and_prepare_role(argv.pop(0))
        resource1 = argv.pop(0)
    elif argv[1] == "with":
        resource1 = argv.pop(0)
    else:
        raise CmdLineInputError()

    if argv.pop(0) != "with":
        raise CmdLineInputError()
    if "with" in argv:
        raise CmdLineInputError(
            message="Multiple 'with's cannot be specified.",
            hint=(
                "Use the 'pcs constraint colocation set' command if you want "
                "to create a constraint for more than two resources."
            ),
            show_both_usage_and_message=True,
        )

    if not argv:
        raise CmdLineInputError()
    if len(argv) == 1:
        resource2 = argv.pop(0)
    else:
        if utils.is_score_or_opt(argv[1]):
            resource2 = argv.pop(0)
        else:
            role2 = _validate_and_prepare_role(argv.pop(0))
            resource2 = argv.pop(0)

    score, nv_pairs = _parse_score_options(argv)

    _validate_constraint_resource(cib_dom, resource1)
    _validate_constraint_resource(cib_dom, resource2)

    id_in_nvpairs = None
    for name, value in nv_pairs:
        if name == "id":
            id_valid, id_error = utils.validate_xml_id(value, "constraint id")
            if not id_valid:
                utils.err(id_error)
            if utils.does_id_exist(cib_dom, value):
                utils.err(
                    "id '%s' is already in use, please specify another one"
                    % value
                )
            id_in_nvpairs = True
    if not id_in_nvpairs:
        nv_pairs.append(
            (
                "id",
                utils.find_unique_id(
                    cib_dom,
                    "colocation-%s-%s-%s" % (resource1, resource2, score),
                ),
            )
        )

    (dom, constraintsElement) = getCurrentConstraints(cib_dom)

    # If one role is specified, the other should default to "started"
    if role1 != "" and role2 == "":
        role2 = DEFAULT_ROLE
    if role2 != "" and role1 == "":
        role1 = DEFAULT_ROLE
    element = dom.createElement("rsc_colocation")
    element.setAttribute("rsc", resource1)
    element.setAttribute("with-rsc", resource2)
    element.setAttribute("score", score)
    if role1 != "":
        element.setAttribute("rsc-role", role1)
    if role2 != "":
        element.setAttribute("with-rsc-role", role2)
    for nv_pair in nv_pairs:
        element.setAttribute(nv_pair[0], nv_pair[1])
    if not modifiers.get("--force"):
        duplicates = colocation_find_duplicates(constraintsElement, element)
        if duplicates:
            utils.err(
                "duplicate constraint already exists, use --force to override\n"
                + "\n".join(
                    [
                        "  "
                        + colocation_format.constraint_plain(
                            {"options": dict(dup.attributes.items())}, True
                        )
                        for dup in duplicates
                    ]
                )
            )
    constraintsElement.appendChild(element)
    utils.replace_cib_configuration(dom)


def colocation_find_duplicates(dom, constraint_el):
    """
    Commandline options: no options
    """
    new_roles_supported = utils.isCibVersionSatisfied(
        dom, const.PCMK_NEW_ROLES_CIB_VERSION
    )

    def normalize(const_el):
        return (
            const_el.getAttribute("rsc"),
            const_el.getAttribute("with-rsc"),
            pacemaker.role.get_value_for_cib(
                const_el.getAttribute("rsc-role").capitalize() or DEFAULT_ROLE,
                new_roles_supported,
            ),
            pacemaker.role.get_value_for_cib(
                const_el.getAttribute("with-rsc-role").capitalize()
                or DEFAULT_ROLE,
                new_roles_supported,
            ),
        )

    normalized_el = normalize(constraint_el)
    return [
        other_el
        for other_el in dom.getElementsByTagName("rsc_colocation")
        if not other_el.getElementsByTagName("resource_set")
        and constraint_el is not other_el
        and normalized_el == normalize(other_el)
    ]


def order_rm(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    elementFound = False
    (dom, constraintsElement) = getCurrentConstraints()

    for resource in argv:
        for ord_loc in constraintsElement.getElementsByTagName("rsc_order")[:]:
            if (
                ord_loc.getAttribute("first") == resource
                or ord_loc.getAttribute("then") == resource
            ):
                constraintsElement.removeChild(ord_loc)
                elementFound = True

        resource_refs_to_remove = []
        for ord_set in constraintsElement.getElementsByTagName("resource_ref"):
            if ord_set.getAttribute("id") == resource:
                resource_refs_to_remove.append(ord_set)
                elementFound = True

        for res_ref in resource_refs_to_remove:
            res_set = res_ref.parentNode
            res_order = res_set.parentNode

            res_ref.parentNode.removeChild(res_ref)
            if not res_set.getElementsByTagName("resource_ref"):
                res_set.parentNode.removeChild(res_set)
                if not res_order.getElementsByTagName("resource_set"):
                    res_order.parentNode.removeChild(res_order)

    if elementFound:
        utils.replace_cib_configuration(dom)
    else:
        utils.err("No matching resources found in ordering list")


def order_start(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --force - allow constraint for any resource, allow duplicate constraints
    """
    del lib
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    first_action = DEFAULT_ACTION
    then_action = DEFAULT_ACTION
    action = argv[0]
    if action in OPTIONS_ACTION:
        first_action = action
        argv.pop(0)

    resource1 = argv.pop(0)
    if argv.pop(0) != "then":
        raise CmdLineInputError()

    if not argv:
        raise CmdLineInputError()

    action = argv[0]
    if action in OPTIONS_ACTION:
        then_action = action
        argv.pop(0)

    if not argv:
        raise CmdLineInputError()
    resource2 = argv.pop(0)

    order_options = []
    if argv:
        order_options = order_options + argv[:]
    if "then" in order_options:
        raise CmdLineInputError(
            message="Multiple 'then's cannot be specified.",
            hint=(
                "Use the 'pcs constraint order set' command if you want to "
                "create a constraint for more than two resources."
            ),
            show_both_usage_and_message=True,
        )

    order_options.append("first-action=" + first_action)
    order_options.append("then-action=" + then_action)
    _order_add(resource1, resource2, order_options, modifiers)


def _order_add(resource1, resource2, options_list, modifiers):
    """
    Commandline options:
      * -f - CIB file
      * --force - allow constraint for any resource, allow duplicate constraints
    """
    cib_dom = utils.get_cib_dom()
    _validate_constraint_resource(cib_dom, resource1)
    _validate_constraint_resource(cib_dom, resource2)

    _validate_resources_not_in_same_group(cib_dom, resource1, resource2)

    order_options = []
    id_specified = False
    sym = None
    for arg in options_list:
        if arg == "symmetrical":
            sym = "true"
        elif arg == "nonsymmetrical":
            sym = "false"
        else:
            name, value = parse_args.split_option(arg, allow_empty_value=False)
            if name == "id":
                id_valid, id_error = utils.validate_xml_id(
                    value, "constraint id"
                )
                if not id_valid:
                    utils.err(id_error)
                if utils.does_id_exist(cib_dom, value):
                    utils.err(
                        "id '%s' is already in use, please specify another one"
                        % value
                    )
                id_specified = True
                order_options.append((name, value))
            elif name == "symmetrical":
                if value.lower() in OPTIONS_SYMMETRICAL:
                    sym = value.lower()
                else:
                    utils.err(
                        "invalid symmetrical value '%s', allowed values are: %s"
                        % (value, ", ".join(OPTIONS_SYMMETRICAL))
                    )
            else:
                order_options.append((name, value))
    if sym:
        order_options.append(("symmetrical", sym))

    options = ""
    if order_options:
        options = " (Options: %s)" % " ".join(
            [
                "%s=%s" % (name, value)
                for name, value in order_options
                if name not in ("kind", "score")
            ]
        )

    scorekind = "kind: Mandatory"
    id_suffix = "mandatory"
    for opt in order_options:
        if opt[0] == "score":
            scorekind = "score: " + opt[1]
            id_suffix = opt[1]
            break
        if opt[0] == "kind":
            scorekind = "kind: " + opt[1]
            id_suffix = opt[1]
            break

    if not id_specified:
        order_id = "order-" + resource1 + "-" + resource2 + "-" + id_suffix
        order_id = utils.find_unique_id(cib_dom, order_id)
        order_options.append(("id", order_id))

    (dom, constraintsElement) = getCurrentConstraints()
    element = dom.createElement("rsc_order")
    element.setAttribute("first", resource1)
    element.setAttribute("then", resource2)
    for order_opt in order_options:
        element.setAttribute(order_opt[0], order_opt[1])
    constraintsElement.appendChild(element)
    if not modifiers.get("--force"):
        duplicates = order_find_duplicates(constraintsElement, element)
        if duplicates:
            utils.err(
                "duplicate constraint already exists, use --force to override\n"
                + "\n".join(
                    [
                        "  "
                        + order_format.constraint_plain(
                            {"options": dict(dup.attributes.items())}, True
                        )
                        for dup in duplicates
                    ]
                )
            )
    print(
        "Adding "
        + resource1
        + " "
        + resource2
        + " ("
        + scorekind
        + ")"
        + options
    )

    utils.replace_cib_configuration(dom)


def order_find_duplicates(dom, constraint_el):
    """
    Commandline options: no options
    """

    def normalize(constraint_el):
        return (
            constraint_el.getAttribute("first"),
            constraint_el.getAttribute("then"),
            constraint_el.getAttribute("first-action").lower()
            or DEFAULT_ACTION,
            constraint_el.getAttribute("then-action").lower() or DEFAULT_ACTION,
        )

    normalized_el = normalize(constraint_el)
    return [
        other_el
        for other_el in dom.getElementsByTagName("rsc_order")
        if not other_el.getElementsByTagName("resource_set")
        and constraint_el is not other_el
        and normalized_el == normalize(other_el)
    ]


def location_show(lib, argv, modifiers):
    warn(
        "This command is deprecated and will be removed. "
        "Please use 'pcs constraint location config' instead.",
        stderr=True,
    )
    return location_config_cmd(lib, argv, modifiers)


# Show the currently configured location constraints by node or resource
def location_config_cmd(lib, argv, modifiers):
    """
    Options:
      * --all - print expired constraints
      * --full - print all details
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f", "--full", "--all")
    by_node = False

    if argv and argv[0] == "nodes":
        by_node = True

    if len(argv) > 1:
        if by_node:
            valid_noderes = argv[1:]
        else:
            valid_noderes = [
                parse_args.parse_typed_arg(
                    arg,
                    [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
                    RESOURCE_TYPE_RESOURCE,
                )
                for arg in argv[1:]
            ]
    else:
        valid_noderes = []

    (dummy_dom, constraintsElement) = getCurrentConstraints()
    print(
        "\n".join(
            location_lines(
                constraintsElement,
                showDetail=modifiers.get("--full"),
                byNode=by_node,
                valid_noderes=valid_noderes,
                show_expired=modifiers.get("--all"),
            )
        )
    )


def location_lines(
    constraintsElement,
    showDetail=False,
    byNode=False,
    valid_noderes=None,
    show_expired=False,
    verify_expiration=True,
):
    """
    Commandline options: no options
    """
    all_lines = []
    nodehashon = {}
    nodehashoff = {}
    rschashon = {}
    rschashoff = {}
    ruleshash = defaultdict(list)
    all_loc_constraints = constraintsElement.getElementsByTagName(
        "rsc_location"
    )
    cib = utils.get_cib()

    if not isfile(settings.crm_rule):
        if verify_expiration:
            sys.stderr.write(CRM_RULE_MISSING_MSG)
        verify_expiration = False

    all_lines.append("Location Constraints:")
    for rsc_loc in all_loc_constraints:
        if rsc_loc.hasAttribute("rsc-pattern"):
            lc_rsc_type = RESOURCE_TYPE_REGEXP
            lc_rsc_value = rsc_loc.getAttribute("rsc-pattern")
            lc_name = "Resource pattern: {0}".format(lc_rsc_value)
        else:
            lc_rsc_type = RESOURCE_TYPE_RESOURCE
            lc_rsc_value = rsc_loc.getAttribute("rsc")
            lc_name = "Resource: {0}".format(lc_rsc_value)
        lc_rsc = lc_rsc_type, lc_rsc_value, lc_name
        lc_id = rsc_loc.getAttribute("id")
        lc_node = rsc_loc.getAttribute("node")
        lc_score = rsc_loc.getAttribute("score")
        lc_role = rsc_loc.getAttribute("role")
        lc_resource_discovery = rsc_loc.getAttribute("resource-discovery")

        for child in rsc_loc.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.tagName == "rule":
                ruleshash[lc_rsc].append(child)

        # NEED TO FIX FOR GROUP LOCATION CONSTRAINTS (where there are children
        # of # rsc_location)
        if lc_score == "":
            lc_score = "0"

        if lc_score == "INFINITY":
            positive = True
        elif lc_score == "-INFINITY":
            positive = False
        elif int(lc_score) >= 0:
            positive = True
        else:
            positive = False

        if positive:
            nodeshash = nodehashon
            rschash = rschashon
        else:
            nodeshash = nodehashoff
            rschash = rschashoff

        hash_element = {
            "id": lc_id,
            "rsc_type": lc_rsc_type,
            "rsc_value": lc_rsc_value,
            "rsc_label": lc_name,
            "node": lc_node,
            "score": lc_score,
            "role": lc_role,
            "resource-discovery": lc_resource_discovery,
        }
        if lc_node in nodeshash:
            nodeshash[lc_node].append(hash_element)
        else:
            nodeshash[lc_node] = [hash_element]
        if lc_rsc in rschash:
            rschash[lc_rsc].append(hash_element)
        else:
            rschash[lc_rsc] = [hash_element]

    nodelist = sorted(set(list(nodehashon.keys()) + list(nodehashoff.keys())))
    rsclist = sorted(
        set(list(rschashon.keys()) + list(rschashoff.keys())),
        key=lambda item: (
            {
                RESOURCE_TYPE_RESOURCE: 1,
                RESOURCE_TYPE_REGEXP: 0,
            }[item[0]],
            item[1],
        ),
    )

    if byNode:
        for node in nodelist:
            if valid_noderes:
                if node not in valid_noderes:
                    continue
            all_lines.append("  Node: " + node)

            nodehash_label = (
                (nodehashon, "    Allowed to run:"),
                (nodehashoff, "    Not allowed to run:"),
            )
            all_lines += _hashtable_to_lines(
                nodehash_label, "rsc_label", node, showDetail
            )
        all_lines += _show_location_rules(
            ruleshash,
            cib,
            show_detail=showDetail,
            show_expired=show_expired,
            verify_expiration=verify_expiration,
        )
    else:
        for rsc in rsclist:
            rsc_lines = []
            if valid_noderes:
                if rsc[0:2] not in valid_noderes:
                    continue
            rsc_lines.append("  {0}".format(rsc[2]))
            rschash_label = (
                (rschashon, "    Enabled on:"),
                (rschashoff, "    Disabled on:"),
            )
            rsc_lines += _hashtable_to_lines(
                rschash_label, "node", rsc, showDetail
            )
            miniruleshash = {}
            miniruleshash[rsc] = ruleshash[rsc]
            rsc_lines += _show_location_rules(
                miniruleshash,
                cib,
                show_detail=showDetail,
                show_expired=show_expired,
                verify_expiration=verify_expiration,
                noheader=True,
            )
            # Append to all_lines only if the resource has any constraints
            if len(rsc_lines) > 2:
                all_lines += rsc_lines
    return all_lines


def _hashtable_to_lines(hash_label, hash_type, hash_key, show_detail):
    hash_lines = []
    for hashtable, label in hash_label:
        if hash_key in hashtable:
            labeled_lines = []
            for options in hashtable[hash_key]:
                # Skips nodeless constraints and prints nodes/resources
                if not options[hash_type]:
                    continue
                line_parts = [
                    "      {0}{1}".format(
                        "Node: " if hash_type == "node" else "",
                        options[hash_type],
                    )
                ]
                line_parts.append(f"(score:{options['score']})")
                if options["role"]:
                    line_parts.append(f"(role:{options['role']})")
                if options["resource-discovery"]:
                    line_parts.append(
                        "(resource-discovery={0})".format(
                            options["resource-discovery"]
                        )
                    )
                if show_detail:
                    line_parts.append(f"(id:{options['id']})")
                labeled_lines.append(" ".join(line_parts))
            if labeled_lines:
                labeled_lines.insert(0, label)
            hash_lines += labeled_lines
    return hash_lines


def _show_location_rules(
    ruleshash,
    cib,
    show_detail,
    show_expired=False,
    verify_expiration=True,
    noheader=False,
):
    """
    Commandline options: no options
    """
    all_lines = []
    constraint_options = {}
    for rsc in sorted(
        ruleshash.keys(),
        key=lambda item: (
            {
                RESOURCE_TYPE_RESOURCE: 1,
                RESOURCE_TYPE_REGEXP: 0,
            }[item[0]],
            item[1],
        ),
    ):
        constrainthash = defaultdict(list)
        if not noheader:
            all_lines.append("  {0}".format(rsc[2]))
        for rule in ruleshash[rsc]:
            constraint_id = rule.parentNode.getAttribute("id")
            constrainthash[constraint_id].append(rule)
            constraint_options[constraint_id] = []
            if rule.parentNode.getAttribute("resource-discovery"):
                constraint_options[constraint_id].append(
                    "resource-discovery=%s"
                    % rule.parentNode.getAttribute("resource-discovery")
                )

        for constraint_id in sorted(constrainthash.keys()):
            if (
                constraint_id in constraint_options
                and constraint_options[constraint_id]
            ):
                constraint_option_info = (
                    " (" + " ".join(constraint_options[constraint_id]) + ")"
                )
            else:
                constraint_option_info = ""

            rule_lines = []
            # When expiration check is needed, starting value should be True and
            # when it's not, check is skipped so the initial value must be False
            # to print the constraint
            is_constraint_expired = verify_expiration
            for rule in constrainthash[constraint_id]:
                rule_status = RULE_UNKNOWN_STATUS
                if verify_expiration:
                    rule_status = _get_rule_status(rule.getAttribute("id"), cib)
                    if rule_status != RULE_EXPIRED:
                        is_constraint_expired = False

                rule_lines.append(
                    rule_utils.ExportDetailed().get_string(
                        rule,
                        rule_status == RULE_EXPIRED and show_expired,
                        show_detail,
                        indent="      ",
                    )
                )

            if not show_expired and is_constraint_expired:
                continue

            all_lines.append(
                "    Constraint{0}: {1}{2}".format(
                    " (expired)" if is_constraint_expired else "",
                    constraint_id,
                    constraint_option_info,
                )
            )
            all_lines += rule_lines
    return all_lines


def _verify_node_name(node, existing_nodes):
    report_list = []
    if node not in existing_nodes:
        report_list.append(
            ReportItem.error(
                reports.messages.NodeNotFound(node),
                force_code=reports.codes.FORCE,
            )
        )
    return report_list


def _verify_score(score):
    if not utils.is_score(score):
        utils.err(
            "invalid score '%s', use integer or INFINITY or -INFINITY" % score
        )


def _get_rule_status(rule_id, cib):
    _, _, retval = utils.cmd_runner().run(
        [settings.crm_rule, "--check", "--rule", rule_id, "--xml-text", "-"],
        cib,
    )
    translation_map = {
        CrmRuleReturnCode.IN_EFFECT.value: RULE_IN_EFFECT,
        CrmRuleReturnCode.EXPIRED.value: RULE_EXPIRED,
        CrmRuleReturnCode.TO_BE_IN_EFFECT.value: RULE_NOT_IN_EFFECT,
    }
    return translation_map.get(retval, RULE_UNKNOWN_STATUS)


def location_prefer(lib, argv, modifiers):
    """
    Options:
      * --force - allow unknown options, allow constraint for any resource type
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    rsc = argv.pop(0)
    prefer_option = argv.pop(0)

    dummy_rsc_type, rsc_value = parse_args.parse_typed_arg(
        rsc,
        [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
        RESOURCE_TYPE_RESOURCE,
    )

    if prefer_option == "prefers":
        prefer = True
    elif prefer_option == "avoids":
        prefer = False
    else:
        raise CmdLineInputError()

    skip_node_check = False
    if modifiers.is_specified("-f") or modifiers.get("--force"):
        skip_node_check = True
        warn(LOCATION_NODE_VALIDATION_SKIP_MSG)
    else:
        lib_env = utils.get_lib_env()
        existing_nodes, report_list = get_existing_nodes_names(
            corosync_conf=lib_env.get_corosync_conf(),
            cib=lib_env.get_cib(),
        )
        if report_list:
            process_library_reports(report_list)

    report_list = []
    parameters_list = []
    for nodeconf in argv:
        nodeconf_a = nodeconf.split("=", 1)
        node = nodeconf_a[0]
        if not skip_node_check:
            report_list += _verify_node_name(node, existing_nodes)
        if len(nodeconf_a) == 1:
            if prefer:
                score = "INFINITY"
            else:
                score = "-INFINITY"
        else:
            score = nodeconf_a[1]
            _verify_score(score)
            if not prefer:
                if score[0] == "-":
                    score = score[1:]
                else:
                    score = "-" + score

        parameters_list.append(
            [
                sanitize_id(f"location-{rsc_value}-{node}-{score}"),
                rsc,
                node,
                score,
            ]
        )

    if report_list:
        process_library_reports(report_list)

    modifiers = modifiers.get_subset("--force", "-f")

    for parameters in parameters_list:
        location_add(lib, parameters, modifiers, skip_score_and_node_check=True)


def location_add(lib, argv, modifiers, skip_score_and_node_check=False):
    """
    Options:
      * --force - allow unknown options, allow constraint for any resource type
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("--force", "-f")
    if len(argv) < 4:
        raise CmdLineInputError()

    constraint_id = argv.pop(0)
    rsc_type, rsc_value = parse_args.parse_typed_arg(
        argv.pop(0),
        [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
        RESOURCE_TYPE_RESOURCE,
    )
    node = argv.pop(0)
    score = argv.pop(0)
    options = []
    # For now we only allow setting resource-discovery
    if argv:
        for arg in argv:
            if "=" in arg:
                options.append(arg.split("=", 1))
            else:
                raise CmdLineInputError(f"bad option '{arg}'")
            if options[-1][0] != "resource-discovery" and not modifiers.get(
                "--force"
            ):
                utils.err(
                    "bad option '%s', use --force to override" % options[-1][0]
                )

    # Verify that specified node exists in the cluster and score is valid
    if not skip_score_and_node_check:
        if modifiers.is_specified("-f") or modifiers.get("--force"):
            warn(LOCATION_NODE_VALIDATION_SKIP_MSG)
        else:
            lib_env = utils.get_lib_env()
            existing_nodes, report_list = get_existing_nodes_names(
                corosync_conf=lib_env.get_corosync_conf(),
                cib=lib_env.get_cib(),
            )
            report_list += _verify_node_name(node, existing_nodes)
            if report_list:
                process_library_reports(report_list)
        _verify_score(score)

    id_valid, id_error = utils.validate_xml_id(constraint_id, "constraint id")
    if not id_valid:
        utils.err(id_error)

    dom = utils.get_cib_dom()

    if rsc_type == RESOURCE_TYPE_RESOURCE:
        (
            rsc_valid,
            rsc_error,
            dummy_correct_id,
        ) = utils.validate_constraint_resource(dom, rsc_value)
        if not rsc_valid:
            utils.err(rsc_error)

    # Verify current constraint doesn't already exist
    # If it does we replace it with the new constraint
    dummy_dom, constraintsElement = getCurrentConstraints(dom)
    elementsToRemove = []
    # If the id matches, or the rsc & node match, then we replace/remove
    for rsc_loc in constraintsElement.getElementsByTagName("rsc_location"):
        # pylint: disable=too-many-boolean-expressions
        if rsc_loc.getAttribute("id") == constraint_id or (
            rsc_loc.getAttribute("node") == node
            and (
                (
                    RESOURCE_TYPE_RESOURCE == rsc_type
                    and rsc_loc.getAttribute("rsc") == rsc_value
                )
                or (
                    RESOURCE_TYPE_REGEXP == rsc_type
                    and rsc_loc.getAttribute("rsc-pattern") == rsc_value
                )
            )
        ):
            elementsToRemove.append(rsc_loc)
    for etr in elementsToRemove:
        constraintsElement.removeChild(etr)

    element = dom.createElement("rsc_location")
    element.setAttribute("id", constraint_id)
    if rsc_type == RESOURCE_TYPE_RESOURCE:
        element.setAttribute("rsc", rsc_value)
    elif rsc_type == RESOURCE_TYPE_REGEXP:
        element.setAttribute("rsc-pattern", rsc_value)
    element.setAttribute("node", node)
    element.setAttribute("score", score)
    for option in options:
        element.setAttribute(option[0], option[1])
    constraintsElement.appendChild(element)

    utils.replace_cib_configuration(dom)


def location_remove(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    # This code was originally merged in the location_add function and was
    # documented to take 1 or 4 arguments:
    # location remove <id> [<resource id> <node> <score>]
    # However it has always ignored all arguments but constraint id. Therefore
    # this command / function has no use as it can be fully replaced by "pcs
    # constraint remove" which also removes constraints by id. For now I keep
    # things as they are but we should solve this when moving these functions
    # to pcs.lib.
    del lib
    modifiers.ensure_only_supported("-f")
    if len(argv) != 1:
        raise CmdLineInputError()

    constraint_id = argv.pop(0)
    dom, constraintsElement = getCurrentConstraints()

    elementsToRemove = []
    for rsc_loc in constraintsElement.getElementsByTagName("rsc_location"):
        if constraint_id == rsc_loc.getAttribute("id"):
            elementsToRemove.append(rsc_loc)

    if not elementsToRemove:
        utils.err("resource location id: " + constraint_id + " not found.")
    for etr in elementsToRemove:
        constraintsElement.removeChild(etr)

    utils.replace_cib_configuration(dom)


def location_rule(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --force - allow constraint on any resource type, allow duplicate
        constraints
    """
    del lib
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        usage.constraint(["location", "rule"])
        sys.exit(1)

    rsc_type, rsc_value = parse_args.parse_typed_arg(
        argv.pop(0),
        [RESOURCE_TYPE_RESOURCE, RESOURCE_TYPE_REGEXP],
        RESOURCE_TYPE_RESOURCE,
    )
    argv.pop(0)  # pop "rule"
    options, rule_argv = rule_utils.parse_argv(
        argv,
        {
            "constraint-id": None,
            "resource-discovery": None,
        },
    )
    resource_discovery = (
        "resource-discovery" in options and options["resource-discovery"]
    )

    try:
        # Parse the rule to see if we need to upgrade CIB schema. All errors
        # would be properly reported by a validator called bellow, so we can
        # safely ignore them here.
        parsed_rule = rule_utils.RuleParser().parse(
            rule_utils.TokenPreprocessor().run(rule_argv)
        )
        if rule_utils.has_node_attr_expr_with_type_integer(parsed_rule):
            utils.checkAndUpgradeCIB(
                const.PCMK_RULES_NODE_ATTR_EXPR_WITH_INT_TYPE_CIB_VERSION
            )
    except (rule_utils.ParserException, rule_utils.CibBuilderException):
        pass

    dom = utils.get_cib_dom()

    if rsc_type == RESOURCE_TYPE_RESOURCE:
        (
            rsc_valid,
            rsc_error,
            dummy_correct_id,
        ) = utils.validate_constraint_resource(dom, rsc_value)
        if not rsc_valid:
            utils.err(rsc_error)

    cib, constraints = getCurrentConstraints(dom)
    lc = cib.createElement("rsc_location")

    # If resource-discovery is specified, we use it with the rsc_location
    # element not the rule
    if resource_discovery:
        lc.setAttribute("resource-discovery", options.pop("resource-discovery"))

    constraints.appendChild(lc)
    if options.get("constraint-id"):
        id_valid, id_error = utils.validate_xml_id(
            options["constraint-id"], "constraint id"
        )
        if not id_valid:
            utils.err(id_error)
        if utils.does_id_exist(dom, options["constraint-id"]):
            utils.err(
                "id '%s' is already in use, please specify another one"
                % options["constraint-id"]
            )
        lc.setAttribute("id", options["constraint-id"])
        del options["constraint-id"]
    else:
        lc.setAttribute(
            "id",
            utils.find_unique_id(dom, sanitize_id("location-" + rsc_value)),
        )
    if rsc_type == RESOURCE_TYPE_RESOURCE:
        lc.setAttribute("rsc", rsc_value)
    elif rsc_type == RESOURCE_TYPE_REGEXP:
        lc.setAttribute("rsc-pattern", rsc_value)

    rule_utils.dom_rule_add(
        lc, options, rule_argv, utils.getValidateWithVersion(cib)
    )
    location_rule_check_duplicates(constraints, lc, modifiers.get("--force"))
    utils.replace_cib_configuration(cib)


def location_rule_check_duplicates(dom, constraint_el, force):
    """
    Commandline options: no options
    """
    if not force:
        duplicates = location_rule_find_duplicates(dom, constraint_el)
        if duplicates:
            lines = []
            for dup in duplicates:
                lines.append("  Constraint: %s" % dup.getAttribute("id"))
                for dup_rule in utils.dom_get_children_by_tag_name(dup, "rule"):
                    lines.append(
                        rule_utils.ExportDetailed().get_string(
                            dup_rule, False, True, indent="    "
                        )
                    )
            utils.err(
                "duplicate constraint already exists, use --force to override\n"
                + "\n".join(lines)
            )


def location_rule_find_duplicates(dom, constraint_el):
    """
    Commandline options: no options
    """

    def normalize(constraint_el):
        if constraint_el.hasAttribute("rsc-pattern"):
            rsc = (
                RESOURCE_TYPE_REGEXP,
                constraint_el.getAttribute("rsc-pattern"),
            )
        else:
            rsc = (RESOURCE_TYPE_RESOURCE, constraint_el.getAttribute("rsc"))
        return (
            rsc,
            [
                rule_utils.ExportAsExpression().get_string(rule_el, True)
                for rule_el in constraint_el.getElementsByTagName("rule")
            ],
        )

    normalized_el = normalize(constraint_el)
    return [
        other_el
        for other_el in dom.getElementsByTagName("rsc_location")
        if other_el.getElementsByTagName("rule")
        and constraint_el is not other_el
        and normalized_el == normalize(other_el)
    ]


# Grabs the current constraints and returns the dom and constraint element
def getCurrentConstraints(passed_dom=None):
    """
    Commandline options:
      * -f - CIB file, only if passed_dom is None
    """
    if passed_dom:
        dom = passed_dom
    else:
        current_constraints_xml = utils.get_cib_xpath("//constraints")
        if current_constraints_xml == "":
            utils.err("unable to process cib")
        # Verify current constraint doesn't already exist
        # If it does we replace it with the new constraint
        dom = parseString(current_constraints_xml)

    constraintsElement = dom.getElementsByTagName("constraints")[0]
    return (dom, constraintsElement)


# If returnStatus is set, then we don't error out, we just print the error
# and return false
def constraint_rm(
    lib,
    argv,
    modifiers,
    returnStatus=False,
    constraintsElement=None,
    passed_dom=None,
):
    """
    Options:
      * -f - CIB file, effective only if passed_dom is None
    """
    if passed_dom is None:
        modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    bad_constraint = False
    if len(argv) != 1:
        for arg in argv:
            if not constraint_rm(
                lib, [arg], modifiers, returnStatus=True, passed_dom=passed_dom
            ):
                bad_constraint = True
        if bad_constraint:
            sys.exit(1)
        return None

    c_id = argv.pop(0)
    elementFound = False

    if not constraintsElement:
        (dom, constraintsElement) = getCurrentConstraints(passed_dom)
        use_cibadmin = True
    else:
        use_cibadmin = False

    for co in constraintsElement.childNodes[:]:
        if co.nodeType != xml.dom.Node.ELEMENT_NODE:
            continue
        if co.getAttribute("id") == c_id:
            constraintsElement.removeChild(co)
            elementFound = True

    if not elementFound:
        for rule in constraintsElement.getElementsByTagName("rule")[:]:
            if rule.getAttribute("id") == c_id:
                elementFound = True
                parent = rule.parentNode
                parent.removeChild(rule)
                if not parent.getElementsByTagName("rule"):
                    parent.parentNode.removeChild(parent)

    if elementFound:
        if passed_dom:
            return dom
        if use_cibadmin:
            utils.replace_cib_configuration(dom)
        if returnStatus:
            return True
    else:
        utils.err("Unable to find constraint - '%s'" % c_id, False)
        if returnStatus:
            return False
        sys.exit(1)
    return None


def constraint_ref(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
    """
    del lib
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    for arg in argv:
        print("Resource: %s" % arg)
        constraints, set_constraints = find_constraints_containing(arg)
        if not constraints and not set_constraints:
            print("  No Matches.")
        else:
            for constraint in constraints:
                print("  " + constraint)
            for constraint in sorted(set_constraints):
                print("  " + constraint)


def remove_constraints_containing(
    resource_id, output=False, constraints_element=None, passed_dom=None
):
    """
    Commandline options:
      * -f - CIB file, effective only if passed_dom is None
    """
    lib = utils.get_library_wrapper()
    modifiers = utils.get_input_modifiers()
    constraints, set_constraints = find_constraints_containing(
        resource_id, passed_dom
    )
    for c in constraints:
        if output:
            print("Removing Constraint - " + c)
        if constraints_element is not None:
            constraint_rm(
                lib,
                [c],
                modifiers,
                True,
                constraints_element,
                passed_dom=passed_dom,
            )
        else:
            constraint_rm(lib, [c], modifiers, passed_dom=passed_dom)

    if set_constraints:
        (dom, constraintsElement) = getCurrentConstraints(passed_dom)
        for c in constraintsElement.getElementsByTagName("resource_ref")[:]:
            # If resource id is in a set, remove it from the set, if the set
            # is empty, then we remove the set, if the parent of the set
            # is empty then we remove it
            if c.getAttribute("id") == resource_id:
                pn = c.parentNode
                pn.removeChild(c)
                if output:
                    print(
                        "Removing %s from set %s"
                        % (resource_id, pn.getAttribute("id"))
                    )
                if pn.getElementsByTagName("resource_ref").length == 0:
                    print("Removing set %s" % pn.getAttribute("id"))
                    pn2 = pn.parentNode
                    pn2.removeChild(pn)
                    if pn2.getElementsByTagName("resource_set").length == 0:
                        pn2.parentNode.removeChild(pn2)
                        print("Removing constraint %s" % pn2.getAttribute("id"))
        if passed_dom:
            return dom
        utils.replace_cib_configuration(dom)
    return None


def find_constraints_containing(resource_id, passed_dom=None):
    """
    Commandline options:
      * -f - CIB file, effective only if passed_dom is None
    """
    if passed_dom:
        dom = passed_dom
    else:
        dom = utils.get_cib_dom()
    constraints_found = []
    set_constraints = []

    resources = dom.getElementsByTagName("primitive")
    resource_match = None
    for res in resources:
        if res.getAttribute("id") == resource_id:
            resource_match = res
            break

    if resource_match:
        if resource_match.parentNode.tagName in ("master", "clone"):
            constraints_found, set_constraints = find_constraints_containing(
                resource_match.parentNode.getAttribute("id"), dom
            )

    constraints = dom.getElementsByTagName("constraints")
    if not constraints:
        return [], []

    constraints = constraints[0]
    myConstraints = constraints.getElementsByTagName("rsc_colocation")
    myConstraints += constraints.getElementsByTagName("rsc_location")
    myConstraints += constraints.getElementsByTagName("rsc_order")
    myConstraints += constraints.getElementsByTagName("rsc_ticket")
    attr_to_match = ["rsc", "first", "then", "with-rsc", "first", "then"]
    for c in myConstraints:
        for attr in attr_to_match:
            if c.getAttribute(attr) == resource_id:
                constraints_found.append(c.getAttribute("id"))
                break

    setConstraints = constraints.getElementsByTagName("resource_ref")
    for c in setConstraints:
        if c.getAttribute("id") == resource_id:
            set_constraints.append(c.parentNode.parentNode.getAttribute("id"))

    # Remove duplicates
    set_constraints = list(set(set_constraints))
    return constraints_found, set_constraints


def remove_constraints_containing_node(dom, node, output=False):
    """
    Commandline options: no options
    """
    for constraint in find_constraints_containing_node(dom, node):
        if output:
            print("Removing Constraint - %s" % constraint.getAttribute("id"))
        constraint.parentNode.removeChild(constraint)
    return dom


def find_constraints_containing_node(dom, node):
    """
    Commandline options: no options
    """
    return [
        constraint
        for constraint in dom.getElementsByTagName("rsc_location")
        if constraint.getAttribute("node") == node
    ]


# Re-assign any constraints referencing a resource to its parent (a clone
# or master)
def constraint_resource_update(old_id, dom):
    """
    Commandline options: no options
    """
    new_id = None
    clone_ms_parent = utils.dom_get_resource_clone_ms_parent(dom, old_id)
    if clone_ms_parent:
        new_id = clone_ms_parent.getAttribute("id")

    if new_id:
        constraints = dom.getElementsByTagName("rsc_location")
        constraints += dom.getElementsByTagName("rsc_order")
        constraints += dom.getElementsByTagName("rsc_colocation")
        attrs_to_update = ["rsc", "first", "then", "with-rsc"]
        for constraint in constraints:
            for attr in attrs_to_update:
                if constraint.getAttribute(attr) == old_id:
                    constraint.setAttribute(attr, new_id)
    return dom


def constraint_rule(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --force - allow duplicate constraints, only for add command

    NOTE: modifiers check is in subcommand
    """
    del lib
    if len(argv) < 2:
        raise CmdLineInputError()

    found = False
    command = argv.pop(0)

    constraint_id = None

    if command == "add":
        modifiers.ensure_only_supported("-f", "--force")
        constraint_id = argv.pop(0)
        options, rule_argv = rule_utils.parse_argv(argv)
        try:
            # Parse the rule to see if we need to upgrade CIB schema. All errors
            # would be properly reported by a validator called bellow, so we can
            # safely ignore them here.
            parsed_rule = rule_utils.RuleParser().parse(
                rule_utils.TokenPreprocessor().run(rule_argv)
            )
            if rule_utils.has_node_attr_expr_with_type_integer(parsed_rule):
                utils.checkAndUpgradeCIB(
                    const.PCMK_RULES_NODE_ATTR_EXPR_WITH_INT_TYPE_CIB_VERSION
                )
        except (rule_utils.ParserException, rule_utils.CibBuilderException):
            pass
        cib = utils.get_cib_dom()
        constraint = utils.dom_get_element_with_id(
            cib.getElementsByTagName("constraints")[0],
            "rsc_location",
            constraint_id,
        )
        if not constraint:
            utils.err("Unable to find constraint: " + constraint_id)
        rule_utils.dom_rule_add(
            constraint,
            options,
            rule_argv,
            utils.getValidateWithVersion(cib),
        )
        location_rule_check_duplicates(
            cib, constraint, modifiers.get("--force")
        )
        utils.replace_cib_configuration(cib)

    elif command in ["remove", "delete"]:
        modifiers.ensure_only_supported("-f")
        cib = utils.get_cib_etree()
        temp_id = argv.pop(0)
        constraints = cib.find(".//constraints")
        loc_cons = cib.findall(str(".//rsc_location"))

        for loc_con in loc_cons:
            for rule in loc_con:
                if rule.get("id") == temp_id:
                    if len(loc_con) > 1:
                        print("Removing Rule: {0}".format(rule.get("id")))
                        loc_con.remove(rule)
                        found = True
                    else:
                        print(
                            "Removing Constraint: {0}".format(loc_con.get("id"))
                        )
                        constraints.remove(loc_con)
                        found = True
                    break

            if found:
                break

        if found:
            utils.replace_cib_configuration(cib)
        else:
            utils.err("unable to find rule with id: %s" % temp_id)
    else:
        raise CmdLineInputError()



import json
import logging
import sys
from typing import (
    Any,
    Dict,
    Mapping,
    Optional,
)

from dacite import DaciteError

from pcs import (
    settings,
    utils,
)
from pcs.cli.common.env_cli import Env
from pcs.cli.common.lib_wrapper import Library
from pcs.common import communication
from pcs.common.interface import dto
from pcs.common.reports import (
    ReportItem,
    ReportItemList,
    ReportProcessor,
)
from pcs.common.resource_agent.dto import ResourceAgentNameDto
from pcs.lib.errors import LibraryError

SUPPORTED_COMMANDS = {
    "acl.add_permission",
    "acl.assign_role_to_group",
    "acl.assign_role_to_target",
    "acl.create_group",
    "acl.create_role",
    "acl.create_target",
    "acl.remove_group",
    "acl.remove_permission",
    "acl.remove_role",
    "acl.remove_target",
    "acl.unassign_role_from_group",
    "acl.unassign_role_from_target",
    "alert.add_recipient",
    "alert.create_alert",
    "alert.remove_alert",
    "alert.remove_recipient",
    "alert.update_alert",
    "alert.update_recipient",
    "cluster.add_nodes",
    "cluster.generate_cluster_uuid",
    "cluster.node_clear",
    "cluster.remove_nodes",
    "cluster.setup",
    "constraint_colocation.create_with_set",
    "constraint_order.create_with_set",
    "constraint_ticket.create",
    "constraint_ticket.create_with_set",
    "constraint_ticket.remove",
    "fencing_topology.add_level",
    "fencing_topology.remove_all_levels",
    "fencing_topology.remove_levels_by_params",
    "fencing_topology.verify",
    "node.maintenance_unmaintenance_all",
    "node.maintenance_unmaintenance_list",
    "node.standby_unstandby_all",
    "node.standby_unstandby_list",
    "qdevice.client_net_import_certificate",
    "qdevice.qdevice_net_sign_certificate_request",
    "resource_agent.describe_agent",
    "resource_agent.get_agents_list",
    "resource_agent.get_agent_metadata",
    "resource_agent.list_agents",
    "resource_agent.list_agents_for_standard_and_provider",
    "resource_agent.list_ocf_providers",
    "resource_agent.list_standards",
    "resource.ban",
    "resource.create",
    "resource.create_as_clone",
    "resource.create_in_group",
    "resource.disable",
    "resource.disable_safe",
    "resource.disable_simulate",
    "resource.enable",
    "resource.group_add",
    "resource.manage",
    "resource.move",
    "resource.move_autoclean",
    "resource.unmanage",
    "resource.unmove_unban",
    "sbd.disable_sbd",
    "sbd.enable_sbd",
    "services.disable_service",
    "services.enable_service",
    "services.get_services_info",
    "services.start_service",
    "services.stop_service",
    "scsi.unfence_node",
    "scsi.unfence_node_mpath",
    "status.full_cluster_status_plaintext",
    "stonith_agent.describe_agent",
    "stonith_agent.list_agents",
    "stonith.create",
    "stonith.create_in_group",
}


def _convert_input_data(cmd: str, data: Dict[str, Any]) -> Mapping[str, Any]:
    if cmd == "resource_agent.get_agent_metadata":
        try:
            data["agent_name"] = dto.from_dict(
                ResourceAgentNameDto, data["agent_name"]
            )
        except (DaciteError, KeyError) as e:
            _exit(
                communication.const.COM_STATUS_INPUT_ERROR,
                status_msg=str(e),
            )
    return data


def _exit(
    status: communication.types.CommunicationResultStatus,
    status_msg: Optional[str] = None,
    report_list: Optional[ReportItemList] = None,
    data: Any = None,
) -> None:
    json.dump(
        dto.to_dict(
            communication.dto.InternalCommunicationResultDto(
                status,
                status_msg,
                [report.to_dto() for report in (report_list or [])],
                data,
            )
        ),
        sys.stdout,
    )
    sys.exit(0)


def get_cli_env(
    options: communication.dto.InternalCommunicationRequestOptionsDto,
) -> Env:
    env = Env()
    env.user, env.groups = utils.get_cib_user_groups()
    env.known_hosts_getter = utils.read_known_hosts_file
    # Debug messages always go to the processor. The parameter only affects if
    # they will be printed to stdout. We are not printing the messages. Instead
    # we get all the messages the processor got. So the value of the parameter
    # does not matter.
    env.report_processor = LibraryReportProcessor()
    env.request_timeout = (
        options.request_timeout or settings.default_request_timeout
    )
    return env


class LibraryReportProcessor(ReportProcessor):
    processed_items: ReportItemList = []

    def _do_report(self, report_item: ReportItem) -> None:
        self.processed_items.append(report_item)


def main() -> None:
    # pylint: disable=broad-except
    argv = sys.argv[1:]
    if argv:
        _exit(
            communication.const.COM_STATUS_INPUT_ERROR,
            status_msg="No arguments allowed",
        )

    utils.subprocess_setup()
    logging.basicConfig()

    try:
        input_dto = dto.from_dict(
            communication.dto.InternalCommunicationRequestDto,
            json.load(sys.stdin),
        )
        cli_env = get_cli_env(input_dto.options)
        lib = Library(cli_env, utils.get_middleware_factory())
        if input_dto.cmd not in SUPPORTED_COMMANDS:
            _exit(
                communication.const.COM_STATUS_UNKNOWN_CMD,
                status_msg=f"Unknown command '{input_dto.cmd}'",
            )
        for sub_cmd in input_dto.cmd.split("."):
            lib = getattr(lib, sub_cmd)
        output_data = lib(**_convert_input_data(input_dto.cmd, input_dto.cmd_data))  # type: ignore
        _exit(
            communication.const.COM_STATUS_SUCCESS,
            report_list=cli_env.report_processor.processed_items,
            data=(
                dto.to_dict(output_data)
                if isinstance(output_data, dto.DataTransferObject)
                else output_data
            ),
        )
    except LibraryError as e:
        _exit(
            communication.const.COM_STATUS_ERROR,
            report_list=(
                cli_env.report_processor.processed_items + list(e.args)
            ),
        )
    except json.JSONDecodeError as e:
        _exit(
            communication.const.COM_STATUS_INPUT_ERROR,
            status_msg=f"Unable to parse input data: {e.msg}",
        )
    except DaciteError as e:
        _exit(
            communication.const.COM_STATUS_INPUT_ERROR,
            status_msg=str(e),
        )
    except Exception as e:
        # TODO: maybe add traceback?
        _exit(communication.const.COM_STATUS_EXCEPTION, status_msg=str(e))



from pcs import settings
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib import (
    sbd,
    validate,
)
from pcs.lib.communication.nodes import GetOnlineTargets
from pcs.lib.communication.sbd import (
    CheckSbd,
    DisableSbdService,
    EnableSbdService,
    GetSbdConfig,
    GetSbdStatus,
    RemoveStonithWatchdogTimeout,
    SetSbdConfig,
    SetStonithWatchdogTimeoutToZero,
)
from pcs.lib.communication.tools import run as run_com
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.tools import environment_file_to_dict

UNSUPPORTED_SBD_OPTION_LIST = [
    "SBD_WATCHDOG_DEV",
    "SBD_OPTS",
    "SBD_PACEMAKER",
    "SBD_DEVICE",
]
ALLOWED_SBD_OPTION_LIST = [
    "SBD_DELAY_START",
    "SBD_STARTMODE",
    "SBD_WATCHDOG_TIMEOUT",
    "SBD_TIMEOUT_ACTION",
]
_TIMEOUT_ACTION_ALLOWED_VALUES = (
    {"flush", "noflush"},
    {"reboot", "off", "crashdump"},
)
_tuple = lambda set1, set2: {f"{v1},{v2}" for v1 in set1 for v2 in set2}
TIMEOUT_ACTION_ALLOWED_VALUE_LIST = sorted(
    _TIMEOUT_ACTION_ALLOWED_VALUES[0]
    | _TIMEOUT_ACTION_ALLOWED_VALUES[1]
    | _tuple(
        _TIMEOUT_ACTION_ALLOWED_VALUES[0], _TIMEOUT_ACTION_ALLOWED_VALUES[1]
    )
    | _tuple(
        _TIMEOUT_ACTION_ALLOWED_VALUES[1], _TIMEOUT_ACTION_ALLOWED_VALUES[0]
    )
)


def _validate_sbd_options(
    sbd_config, allow_unknown_opts=False, allow_invalid_option_values=False
):
    """
    Validate user SBD configuration. Options 'SBD_WATCHDOG_DEV' and 'SBD_OPTS'
    are restricted. Returns list of ReportItem

    sbd_config -- dictionary in format: <SBD config option>: <value>
    allow_unknown_opts -- if True, accept also unknown options.
    """
    validators = [
        validate.NamesIn(
            ALLOWED_SBD_OPTION_LIST,
            banned_name_list=UNSUPPORTED_SBD_OPTION_LIST,
            severity=reports.item.get_severity(
                reports.codes.FORCE, allow_unknown_opts
            ),
        ),
        validate.ValueNonnegativeInteger("SBD_WATCHDOG_TIMEOUT"),
        validate.ValueIn(
            "SBD_TIMEOUT_ACTION",
            TIMEOUT_ACTION_ALLOWED_VALUE_LIST,
            severity=reports.item.get_severity(
                reports.codes.FORCE, allow_invalid_option_values
            ),
        ),
    ]
    return validate.ValidatorAll(validators).validate(sbd_config)


def _validate_watchdog_dict(watchdog_dict):
    """
    Validates if all watchdogs are not empty strings.
    Returns list of ReportItem.

    watchdog_dict -- dictionary with node names as keys and value as watchdog
    """
    return [
        ReportItem.error(reports.messages.WatchdogInvalid(watchdog))
        for watchdog in watchdog_dict.values()
        if not watchdog
    ]


def _get_full_target_dict(target_list, node_value_dict, default_value):
    """
    Returns dictionary where keys are labels of all nodes in cluster and value
    is obtained from node_value_dict for node name, or default value if node
    is not specified in node_value_dict.

    list node_list -- list of cluster nodes (RequestTarget object)
    node_value_dict -- dictionary, keys: node names, values: some velue
    default_value -- some default value
    """
    return {
        target.label: node_value_dict.get(target.label, default_value)
        for target in target_list
    }


def enable_sbd(
    lib_env,
    default_watchdog,
    watchdog_dict,
    sbd_options,
    default_device_list=None,
    node_device_dict=None,
    allow_unknown_opts=False,
    ignore_offline_nodes=False,
    no_watchdog_validation=False,
    allow_invalid_option_values=False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Enable SBD on all nodes in cluster.

    lib_env -- LibraryEnvironment
    default_watchdog -- watchdog for nodes which are not specified in
        watchdog_dict. Uses default value from settings if None.
    watchdog_dict -- dictionary with node names as keys and watchdog path
        as value
    sbd_options -- dictionary in format: <SBD config option>: <value>
    default_device_list -- list of devices for all nodes
    node_device_dict -- dictionary with node names as keys and list of devices
        as value
    allow_unknown_opts -- if True, accept also unknown options.
    ignore_offline_nodes -- if True, omit offline nodes
    no_watchdog_validation -- it True, do not validate existence of a watchdog
        on the nodes
    allow_invalid_option_values -- if True, invalid values of some options will
        be treated as warning instead of errors
    """
    using_devices = not (
        default_device_list is None and node_device_dict is None
    )
    if default_device_list is None:
        default_device_list = []
    if node_device_dict is None:
        node_device_dict = {}
    if not default_watchdog:
        default_watchdog = settings.sbd_watchdog_default
    sbd_options = {opt.upper(): val for opt, val in sbd_options.items()}

    corosync_conf = lib_env.get_corosync_conf()

    node_list, get_nodes_report_list = get_existing_nodes_names(corosync_conf)
    if not node_list:
        get_nodes_report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    target_list = lib_env.get_node_target_factory().get_target_list(
        node_list,
        skip_non_existing=ignore_offline_nodes,
    )

    full_watchdog_dict = _get_full_target_dict(
        target_list, watchdog_dict, default_watchdog
    )
    full_device_dict = _get_full_target_dict(
        target_list, node_device_dict, default_device_list
    )

    if lib_env.report_processor.report_list(
        get_nodes_report_list
        + [
            ReportItem.error(reports.messages.NodeNotFound(node))
            for node in (
                set(list(watchdog_dict.keys()) + list(node_device_dict.keys()))
                - set(node_list)
            )
        ]
        + _validate_watchdog_dict(full_watchdog_dict)
        + (
            sbd.validate_nodes_devices(full_device_dict)
            if using_devices
            else []
        )
        + _validate_sbd_options(
            sbd_options, allow_unknown_opts, allow_invalid_option_values
        )
    ).has_errors:
        raise LibraryError()

    com_cmd = GetOnlineTargets(
        lib_env.report_processor,
        ignore_offline_targets=ignore_offline_nodes,
    )
    com_cmd.set_targets(target_list)
    online_targets = run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # check if SBD can be enabled
    if no_watchdog_validation:
        lib_env.report_processor.report(
            ReportItem.warning(reports.messages.SbdWatchdogValidationInactive())
        )
    com_cmd = CheckSbd(lib_env.report_processor)
    for target in online_targets:
        com_cmd.add_request(
            target,
            (
                # Do not send watchdog if validation is turned off. Listing of
                # available watchdogs in pcsd may restart the machine in some
                # corner cases.
                ""
                if no_watchdog_validation
                else full_watchdog_dict[target.label]
            ),
            full_device_dict[target.label] if using_devices else [],
        )
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # enable ATB if needed
    if not using_devices:
        if sbd.atb_has_to_be_enabled_pre_enable_check(corosync_conf):
            lib_env.report_processor.report(
                ReportItem.warning(
                    reports.messages.CorosyncQuorumAtbWillBeEnabledDueToSbd()
                )
            )
            corosync_conf.set_quorum_options({"auto_tie_breaker": "1"})
            lib_env.push_corosync_conf(corosync_conf, ignore_offline_nodes)

    # distribute SBD configuration
    config = sbd.get_default_sbd_config()
    config.update(sbd_options)
    com_cmd = SetSbdConfig(lib_env.report_processor)
    for target in online_targets:
        com_cmd.add_request(
            target,
            sbd.create_sbd_config(
                config,
                target.label,
                full_watchdog_dict[target.label],
                full_device_dict[target.label],
            ),
        )
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # remove cluster prop 'stonith_watchdog_timeout'
    com_cmd = RemoveStonithWatchdogTimeout(lib_env.report_processor)
    com_cmd.set_targets(online_targets)
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # enable SBD service an all nodes
    com_cmd = EnableSbdService(lib_env.report_processor)
    com_cmd.set_targets(online_targets)
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    lib_env.report_processor.report(
        ReportItem.warning(
            reports.messages.ClusterRestartRequiredToApplyChanges()
        )
    )


def disable_sbd(lib_env, ignore_offline_nodes=False):
    """
    Disable SBD on all nodes in cluster.

    lib_env -- LibraryEnvironment
    ignore_offline_nodes -- if True, omit offline nodes
    """
    node_list, get_nodes_report_list = get_existing_nodes_names(
        lib_env.get_corosync_conf()
    )
    if not node_list:
        get_nodes_report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    if lib_env.report_processor.report_list(get_nodes_report_list).has_errors:
        raise LibraryError()

    com_cmd = GetOnlineTargets(
        lib_env.report_processor,
        ignore_offline_targets=ignore_offline_nodes,
    )
    com_cmd.set_targets(
        lib_env.get_node_target_factory().get_target_list(
            node_list,
            skip_non_existing=ignore_offline_nodes,
        )
    )
    online_nodes = run_and_raise(lib_env.get_node_communicator(), com_cmd)

    com_cmd = SetStonithWatchdogTimeoutToZero(lib_env.report_processor)
    com_cmd.set_targets(online_nodes)
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    com_cmd = DisableSbdService(lib_env.report_processor)
    com_cmd.set_targets(online_nodes)
    run_and_raise(lib_env.get_node_communicator(), com_cmd)

    lib_env.report_processor.report(
        ReportItem.warning(
            reports.messages.ClusterRestartRequiredToApplyChanges()
        )
    )


def get_cluster_sbd_status(lib_env):
    """
    Returns status of SBD service in cluster in dictionary with format:
    {
        <NodeAddress>: {
            "installed": <boolean>,
            "enabled": <boolean>,
            "running": <boolean>
        },
        ...
    }

    lib_env -- LibraryEnvironment
    """
    node_list, get_nodes_report_list = get_existing_nodes_names(
        lib_env.get_corosync_conf()
    )
    if not node_list:
        get_nodes_report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    if lib_env.report_processor.report_list(get_nodes_report_list).has_errors:
        raise LibraryError()

    com_cmd = GetSbdStatus(lib_env.report_processor)
    com_cmd.set_targets(
        lib_env.get_node_target_factory().get_target_list(
            node_list, skip_non_existing=True
        )
    )
    return run_com(lib_env.get_node_communicator(), com_cmd)


def get_cluster_sbd_config(lib_env):
    """
    Returns list of SBD config from all cluster nodes in cluster. Structure
    of data:
    [
        {
            "node": <NodeAddress>
            "config": <sbd_config_dict> or None if there was failure,
        },
        ...
    ]
    If error occurs while obtaining config from some node, it's config will be
    None. If obtaining config fail on all node returns empty dictionary.

    lib_env -- LibraryEnvironment
    """
    node_list, get_nodes_report_list = get_existing_nodes_names(
        lib_env.get_corosync_conf()
    )
    if not node_list:
        get_nodes_report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    if lib_env.report_processor.report_list(get_nodes_report_list).has_errors:
        raise LibraryError()

    com_cmd = GetSbdConfig(lib_env.report_processor)
    com_cmd.set_targets(
        lib_env.get_node_target_factory().get_target_list(
            node_list, skip_non_existing=True
        )
    )
    return run_com(lib_env.get_node_communicator(), com_cmd)


def get_local_sbd_config(lib_env):
    # pylint: disable=unused-argument
    """
    Returns local SBD config as dictionary.

    lib_env -- LibraryEnvironment
    """
    return environment_file_to_dict(sbd.get_local_sbd_config())


def initialize_block_devices(lib_env, device_list, option_dict):
    """
    Initialize SBD devices in device_list with options_dict.

    lib_env -- LibraryEnvironment
    device_list -- list of strings
    option_dict -- dictionary
    """
    report_item_list = []
    if not device_list:
        report_item_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["device"])
            )
        )

    supported_options = sbd.DEVICE_INITIALIZATION_OPTIONS_MAPPING.keys()

    report_item_list += validate.NamesIn(supported_options).validate(
        option_dict
    )

    report_item_list += validate.ValidatorAll(
        [validate.ValueNonnegativeInteger(key) for key in supported_options]
    ).validate(option_dict)

    if lib_env.report_processor.report_list(report_item_list).has_errors:
        raise LibraryError()
    sbd.initialize_block_devices(
        lib_env.report_processor, lib_env.cmd_runner(), device_list, option_dict
    )


def get_local_devices_info(lib_env, dump=False):
    """
    Returns list of local devices info in format:
    {
        "device": <device_path>,
        "list": <output of 'sbd list' command>,
        "dump": <output of 'sbd dump' command> if dump is True, None otherwise
    }
    If sbd is not enabled, empty list will be returned.

    lib_env -- LibraryEnvironment
    dump -- if True returns also output of command 'sbd dump'
    """
    if not sbd.is_sbd_enabled(lib_env.service_manager):
        return []
    device_list = sbd.get_local_sbd_device_list()
    report_item_list = []
    output = []
    for device in device_list:
        obj = {
            "device": device,
            "list": None,
            "dump": None,
        }
        try:
            obj["list"] = sbd.get_device_messages_info(
                lib_env.cmd_runner(), device
            )
            if dump:
                obj["dump"] = sbd.get_device_sbd_header_dump(
                    lib_env.cmd_runner(), device
                )
        except LibraryError as e:
            report_item_list += e.args

        output.append(obj)

    for report_item in report_item_list:
        report_item.severity = reports.item.ReportItemSeverity.warning()
    if lib_env.report_processor.report_list(report_item_list).has_errors:
        raise LibraryError()
    return output


def set_message(lib_env, device, node_name, message):
    """
    Set message on device for node_name.

    lib_env -- LibraryEnvironment
    device -- string, absolute path to device
    node_name -- string
    message -- string, message type, should be one of settings.sbd_message_types
    """
    report_item_list = []
    missing_options = []
    if not device:
        missing_options.append("device")
    if not node_name:
        missing_options.append("node")
    if missing_options:
        report_item_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(missing_options)
            )
        )
    supported_messages = settings.sbd_message_types
    if message not in supported_messages:
        report_item_list.append(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "message", message, supported_messages
                )
            )
        )
    if lib_env.report_processor.report_list(report_item_list).has_errors:
        raise LibraryError()
    sbd.set_message(lib_env.cmd_runner(), device, node_name, message)


def get_local_available_watchdogs(lib_env):
    """
    Returns available local watchdog devices.

    lib_env LibraryEnvironment
    """
    return sbd.get_available_watchdogs(lib_env.cmd_runner())


def test_local_watchdog(lib_env, watchdog=None):
    """
    Test local watchdog device by triggering it. System reset is expected. If
    watchdog is not specified, available watchdog will be used if there is only
    one.

    lib_env LibraryEnvironment
    watchdog string -- watchdog to trigger
    """
    lib_env.report_processor.report(
        ReportItem.info(reports.messages.SystemWillReset())
    )
    sbd.test_watchdog(lib_env.cmd_runner(), watchdog)
